import numpy as np
import pandas as pd
cimport numpy as np
from scipy.special.cython_special import ndtri, stdtrit, gamma
from collections import deque

from qubx.core.series cimport TimeSeries, Indicator, IndicatorOHLC, RollingSum, nans, OHLCV, Bar, SeriesCachedValue
from qubx.core.utils import time_to_str
from qubx.pandaz.utils import scols, srows


cdef extern from "math.h":
    float INFINITY


cdef inline long long floor_t64(long long time, long long dt):
    """
    Floor timestamp by dt
    """
    return time - time % dt


cdef class Sma(Indicator):
    """
    Simple Moving Average indicator
    """

    def __init__(self, str name, TimeSeries series, int period):
        self.period = period
        self.summator = RollingSum(period)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        cdef double r = self.summator.update(value, new_item_started)
        return np.nan if self.summator.is_init_stage else r / self.period


def sma(series:TimeSeries, period: int): 
    return Sma.wrap(series, period)


cdef class Ema(Indicator):
    """
    Exponential moving average
    """

    def __init__(self, str name, TimeSeries series, int period, init_mean=True):
        self.period = period

        # when it's required to initialize this ema by mean on first period
        self.init_mean = init_mean
        if init_mean:
            self.__s = nans(period)
            self.__i = 0

        self._init_stage = 1
        self.alpha = 2.0 / (1.0 + period)
        self.alpha_1 = (1 - self.alpha)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        cdef int prev_bar_idx = 0 if new_item_started else 1 

        if self._init_stage:
            if np.isnan(value): return np.nan

            if new_item_started:
                self.__i += 1
                if self.__i > self.period - 1:
                    self._init_stage = False
                    return self.alpha * value + self.alpha_1 * self[prev_bar_idx]

            if self.__i == self.period - 1:
                self.__s[self.__i] = value 
                return np.nansum(self.__s) / self.period

            self.__s[self.__i] = value 
            return np.nan

        if len(self) == 0:
            return value

        return self.alpha * value + self.alpha_1 * self[prev_bar_idx]


def ema(series:TimeSeries, period: int, init_mean: bool = True):
    return Ema.wrap(series, period, init_mean=init_mean)


cdef class Tema(Indicator):

    def __init__(self, str name, TimeSeries series, int period, init_mean=True):
        self.period = period
        self.init_mean = init_mean
        self.ser0 = TimeSeries('ser0', series.timeframe, series.max_series_length)
        self.ema1 = ema(self.ser0, period, init_mean)
        self.ema2 = ema(self.ema1, period, init_mean)
        self.ema3 = ema(self.ema2, period, init_mean)
        super().__init__(name, series)
        
    cpdef double calculate(self, long long time, double value, short new_item_started):
        self.ser0.update(time, value)
        return 3 * self.ema1[0] - 3 * self.ema2[0] + self.ema3[0]


def tema(series:TimeSeries, period: int, init_mean: bool = True):
    return Tema.wrap(series, period, init_mean=init_mean)


cdef class Dema(Indicator):

    def __init__(self, str name, TimeSeries series, int period, init_mean=True):
        self.period = period
        self.init_mean = init_mean
        self.ser0 = TimeSeries('ser0', series.timeframe, series.max_series_length)
        self.ema1 = ema(self.ser0, period, init_mean)
        self.ema2 = ema(self.ema1, period, init_mean)
        super().__init__(name, series)
        
    cpdef double calculate(self, long long time, double value, short new_item_started):
        self.ser0.update(time, value)
        return 2 * self.ema1[0] - self.ema2[0]


def dema(series:TimeSeries, period: int, init_mean: bool = True):
    return Dema.wrap(series, period, init_mean=init_mean)


cdef class Kama(Indicator):
    # cdef int period
    # cdef int fast_span
    # cdef int slow_span
    # cdef double _S1 
    # cdef double _K1 
    # cdef _x_past
    # cdef RollingSum summator

    def __init__(self, str name, TimeSeries series, int period, int fast_span=2, int slow_span=30):
        self.period = period
        self.fast_span = fast_span
        self.slow_span = slow_span
        self._S1 = 2.0 / (slow_span + 1)
        self._K1 = 2.0 / (fast_span + 1) - self._S1
        self._x_past = deque(nans(period+1), period+1)
        self.summator = RollingSum(period)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        if new_item_started:
            self._x_past.append(value)
        else:
            self._x_past[-1] = value

        cdef double rs = self.summator.update(abs(value - self._x_past[-2]), new_item_started)
        cdef double er = (abs(value - self._x_past[0]) / rs) if rs != 0.0 else 1.0
        cdef double sc = (er * self._K1 + self._S1) ** 2

        if self.summator.is_init_stage:
            if not np.isnan(self._x_past[1]):
                return value
            return np.nan

        return sc * value + (1 - sc) * self[0 if new_item_started else 1]


def kama(series:TimeSeries, period: int, fast_span:int=2, slow_span:int=30):
    return Kama.wrap(series, period, fast_span, slow_span)


cdef class Highest(Indicator):

    def __init__(self, str name, TimeSeries series, int period):
        self.period = period
        self.queue = deque([np.nan] * period, maxlen=period)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        """
        Not a most effictive algo but simplest and can handle updated last value
        """
        cdef float r = np.nan

        if not np.isnan(value):
            if new_item_started:
                self.queue.append(value)
            else:
                self.queue[-1] = value

        if not np.isnan(self.queue[0]):
            r = max(self.queue) 

        return r


def highest(series:TimeSeries, period:int):
    return Highest.wrap(series, period)


cdef class Lowest(Indicator):

    def __init__(self, str name, TimeSeries series, int period):
        self.period = period
        self.queue = deque([np.nan] * period, maxlen=period)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        """
        Not a most effictive algo but simplest and can handle updated last value
        """
        cdef float r = np.nan

        if not np.isnan(value):
            if new_item_started:
                self.queue.append(value)
            else:
                self.queue[-1] = value

        if not np.isnan(self.queue[0]):
            r = min(self.queue) 

        return r


def lowest(series:TimeSeries, period:int):
    return Lowest.wrap(series, period)


cdef class Std(Indicator):
    """
    Streaming Standard Deviation indicator (uses population std by default)
    Supports min_periods for early estimates like pandas rolling().std()
    """

    def __init__(self, str name, TimeSeries series, int period, int ddof=0, min_periods=None):
        self.period = period
        self.ddof = ddof  # degrees of freedom (0 for population, 1 for sample)
        # If min_periods not specified, use period (default behavior)
        self.min_periods = min_periods if min_periods is not None else period
        # Ensure min_periods is at least ddof + 1 to avoid division by zero
        if self.min_periods < self.ddof + 1:
            self.min_periods = self.ddof + 1

        # Use deque to track values for proper calculation
        self.values_deque = deque(maxlen=period)
        self.count = 0
        self._sum = 0.0
        self._sum_sq = 0.0
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        cdef double old_value
        cdef double _variance
        cdef double _mean
        cdef int n

        if new_item_started:
            # New bar: add value to deque
            if len(self.values_deque) == self.period:
                # Remove oldest value from sums
                old_value = self.values_deque[0]
                if not np.isnan(old_value):
                    self._sum -= old_value
                    self._sum_sq -= old_value * old_value
                    self.count -= 1

            # Add new value
            self.values_deque.append(value)
            if not np.isnan(value):
                self._sum += value
                self._sum_sq += value * value
                self.count += 1
        else:
            # Update current bar: replace last value
            if len(self.values_deque) > 0:
                old_value = self.values_deque[-1]
                if not np.isnan(old_value):
                    self._sum -= old_value
                    self._sum_sq -= old_value * old_value
                    self.count -= 1

                self.values_deque[-1] = value
                if not np.isnan(value):
                    self._sum += value
                    self._sum_sq += value * value
                    self.count += 1

        # Check if we have enough values
        n = len(self.values_deque)
        # We need at least min_periods values in the deque and enough non-NaN values for calculation
        if n < self.min_periods or self.count < max(self.min_periods, self.ddof + 1):
            return np.nan

        # Use actual count of non-NaN values for calculation
        if self.count == 0:
            return np.nan

        _mean = self._sum / self.count

        # Calculate variance
        if self.ddof == 0:
            # Population variance
            _variance = self._sum_sq / self.count - _mean * _mean
        else:
            # Sample variance (Bessel's correction)
            if self.count <= self.ddof:
                return np.nan
            _variance = (self._sum_sq - self._sum * self._sum / self.count) / (self.count - self.ddof)

        # Handle numerical errors that might make variance negative
        if _variance < 0:
            _variance = 0

        # Return the square root of the variance (standard deviation)
        return np.sqrt(_variance)


def std(series: TimeSeries, period: int, ddof: int = 0, min_periods=None):
    return Std.wrap(series, period, ddof, min_periods)


cdef double norm_pdf(double x):
    return np.exp(-x ** 2 / 2) / np.sqrt(2 * np.pi)


cdef double lognorm_pdf(double x, double s):
    return np.exp(-np.log(x) ** 2 / (2 * s ** 2)) / (x * s * np.sqrt(2 * np.pi))


cdef double student_t_pdf(double x, double df):
    """Compute the PDF of the Student's t-distribution."""
    gamma_df = gamma(df / 2.0)
    gamma_df_plus_1 = gamma((df + 1) / 2.0)
    
    # Normalization constant
    normalization = gamma_df_plus_1 / (np.sqrt(df * np.pi) * gamma_df)
    
    # PDF calculation
    term = (1 + (x ** 2) / df) ** (-(df + 1) / 2.0)
    pdf_value = normalization * term
    
    return pdf_value


cdef class Pewma(Indicator):

    def __init__(self, str name, TimeSeries series, double alpha, double beta, int T):
        self.alpha = alpha 
        self.beta = beta
        self.T = T

        # - local variables
        self._i = 0
        self.std = TimeSeries('std', series.timeframe, series.max_series_length)
        super().__init__(name, series)

    def _store(self):
        self.mean = self._mean
        self.vstd = self._vstd
        self.var = self._var

    def _restore(self):
        self._mean = self.mean
        self._vstd = self.vstd
        self._var = self.var

    def _get_alpha(self, p_t):
        if self._i - 1 > self.T:
            return self.alpha * (1.0 - self.beta * p_t)
        return 1.0 - 1.0 / self._i

    cpdef double calculate(self, long long time, double x, short new_item_started):
        cdef double diff, p_t, a_t, incr

        if len(self.series) <= 1:
            self._mean = x
            self._vstd = 0.0
            self._var = 0.0
            self._store()
            self.std.update(time, self.vstd)
            return self.mean

        if new_item_started:
            self._i += 1
            self._restore()
        else:
            self._store()

        diff = x - self.mean
        # prob of observing diff
        p_t = norm_pdf(diff / self.vstd) if self.vstd != 0.0 else 0.0  

        # weight to give to this point
        a_t = self._get_alpha(p_t)  
        incr = (1.0 - a_t) * diff
        self.mean += incr
        self.var = a_t * (self.var + diff * incr)
        self.vstd = np.sqrt(self.var)
        self.std.update(time, self.vstd)

        return self.mean


def pewma(series:TimeSeries, alpha: float, beta: float, T:int=30):
    """
    Implementation of probabilistic exponential weighted ma (https://sci-hub.shop/10.1109/SSP.2012.6319708)
    See pandas version here: qubx.pandaz.ta::pwma 
    """
    return Pewma.wrap(series, alpha, beta, T)


cdef class PewmaOutliersDetector(Indicator):

    def __init__(
        self,
        str name,
        TimeSeries series,
        double alpha,
        double beta,
        int T,
        double threshold,
        str dist = "normal",
        double student_t_df = 3.0
    ):
        self.alpha = alpha 
        self.beta = beta
        self.T = T
        self.threshold = threshold
        self.dist = dist
        self.student_t_df = student_t_df

        # - series
        self.upper = TimeSeries('uba', series.timeframe, series.max_series_length)
        self.lower = TimeSeries('lba', series.timeframe, series.max_series_length)
        self.std = TimeSeries('std', series.timeframe, series.max_series_length)
        self.outliers = TimeSeries('outliers', series.timeframe, series.max_series_length)

        # - local variables
        self._i = 0
        self._z_thr = self._get_z_thr()

        super().__init__(name, series)

    def _store(self):
        self.mean = self._mean
        self.vstd = self._vstd
        self.variance = self._variance

    def _restore(self):
        self._mean = self.mean
        self._vstd = self.vstd
        self._variance = self.variance
    
    cdef double _get_z_thr(self):
        if self.dist == 'normal':
            return ndtri(1 - self.threshold / 2)
        elif self.dist == 'student_t':
            return stdtrit(self.student_t_df, 1 - self.threshold / 2)
        else:
            raise ValueError('Invalid distribution type')

    cdef double _get_alpha(self, double p_t):
        if self._i + 1 >= self.T:
            return self.alpha * (1.0 - self.beta * p_t)
        return 1.0 - 1.0 / (self._i + 1.0)

    cdef double _get_mean(self, double x, double alpha_t):
        return alpha_t * self.mean + (1.0 - alpha_t) * x

    cdef double _get_variance(self, double x, double alpha_t):
        return alpha_t * self.variance + (1.0 - alpha_t) * np.square(x)

    cdef double _get_std(self, double variance, double mean):
        return np.sqrt(max(variance - np.square(mean), 0.0))

    cdef double _get_p(self, double x):
        cdef double z_t = ((x - self.mean) / self.vstd) if (self.vstd != 0 and not np.isnan(x)) else 0.0
        if self.dist == 'normal':
            p_t = norm_pdf(z_t)
        elif self.dist == 'student_t':
            p_t = student_t_pdf(z_t, self.student_t_df)
        # elif self.dist == 'cauchy':
        #     p_t = (1 / (np.pi * (1 + np.square(z_t))))
        else:
            raise ValueError('Invalid distribution type')
        return p_t

    cpdef double calculate(self, long long time, double x, short new_item_started):
        # - first bar - just use it as initial value
        if len(self.series) <= 1:
            self._mean = x
            self._variance = x ** 2
            self._vstd = 0.0
            self._store()
            self.std.update(time, self.vstd)
            self.upper.update(time, x)
            self.lower.update(time, x)
            return self._mean

        # - new bar is started use n-1 values for calculate innovations
        if new_item_started:
            self._i += 1
            self._restore()
        else:
            self._store()

        cdef double p_t = self._get_p(x)
        cdef double alpha_t = self._get_alpha(p_t)
        self.mean = self._get_mean(x, alpha_t)
        self.variance = self._get_variance(x, alpha_t)
        self.vstd = self._get_std(self.variance, self.mean)
        cdef double ub = self.mean + self._z_thr * self.vstd
        cdef double lb = self.mean - self._z_thr * self.vstd

        self.upper.update(time, ub)
        self.lower.update(time, lb)
        self.std.update(time, self.vstd)

        # - check if it's outlier
        if p_t < self.threshold:
            self.outliers.update(time, x)
        else:
            self.outliers.update(time, np.nan)
        return self.mean


def pewma_outliers_detector(
    series: TimeSeries,
    alpha: float,
    beta: float,
    T:int=30,
    threshold=0.05,
    dist: str = "normal",
    **kwargs
):
    """
    Outliers detector based on pwma
    """
    return PewmaOutliersDetector.wrap(series, alpha, beta, T, threshold, dist=dist, **kwargs)


cdef class Psar(IndicatorOHLC):

    def __init__(self, name, series, iaf, maxaf):
        self.iaf = iaf
        self.maxaf = maxaf
        self.upper = TimeSeries('upper', series.timeframe, series.max_series_length)
        self.lower = TimeSeries('lower', series.timeframe, series.max_series_length)
        super().__init__(name, series)

    cdef _store(self):
        self.bull = self._bull
        self.af = self._af
        self.psar = self._psar
        self.lp = self._lp
        self.hp = self._hp

    cdef _restore(self):
        self._bull = self.bull
        self._af = self.af
        self._psar = self.psar
        self._lp = self.lp
        self._hp = self.hp

    cpdef double calculate(self, long long time, Bar bar, short new_item_started):
        cdef short reverse = 1

        if len(self.series) <= 2:
            self._bull = 1
            self._af = self.iaf
            self._psar = bar.close

            if len(self.series) == 1:
                self._lp = bar.low
                self._hp = bar.high
            self._store()
            return self._psar

        if not new_item_started:
            self._store()
        else:
            self._restore()

        bar1 = self.series[1]
        bar2 = self.series[2]
        cdef double h0 = bar.high
        cdef double l0 = bar.low
        cdef double h1 = bar1.high
        cdef double l1 = bar1.low
        cdef double h2 = bar2.high
        cdef double l2 = bar2.low

        if self.bull:
            self.psar += self.af * (self.hp - self.psar)
        else:
            self.psar += self.af * (self.lp - self.psar)

        reverse = 0
        if self.bull:
            if l0 < self.psar:
                self.bull = 0
                reverse = 1
                self.psar = self.hp
                self.lp = l0
                self.af = self.iaf
        else:
            if h0 > self.psar:
                self.bull = 1
                reverse = 1
                self.psar = self.lp
                self.hp = h0
                self.af = self.iaf

        if not reverse:
            if self.bull:
                if h0 > self.hp:
                    self.hp = h0
                    self.af = min(self.af + self.iaf, self.maxaf)
                if l1 < self.psar:
                    self.psar = l1
                if l2 < self.psar:
                    self.psar = l2
            else:
                if l0 < self.lp:
                    self.lp = l0
                    self.af = min(self.af + self.iaf, self.maxaf)
                if h1 > self.psar:
                    self.psar = h1
                if h2 > self.psar:
                    self.psar = h2

        if self.bull:
            self.lower.update(time, self.psar)
            self.upper.update(time, np.nan)
        else:
            self.upper.update(time, self.psar)
            self.lower.update(time, np.nan)

        return self.psar


def psar(series: OHLCV, iaf: float=0.02, maxaf: float=0.2):
    if not isinstance(series, OHLCV):
        raise ValueError('Series must be OHLCV !')

    return Psar.wrap(series, iaf, maxaf)


# List of smoothing functions
_smoothers = {f.__name__: f for f in [pewma, ema, sma, kama, tema, dema]}


def smooth(TimeSeries series, str smoother, *args, **kwargs) -> Indicator:
    """
    Handy utility function to smooth series
    """
    _sfn = _smoothers.get(smoother)
    if _sfn is None:
        raise ValueError(f"Smoother {smoother} not found!")
    return _sfn(series, *args, **kwargs)


cdef class Atr(IndicatorOHLC):

    def __init__(self, str name, OHLCV series, int period, str smoother, short percentage):
        self.percentage = percentage
        self.tr = TimeSeries("tr", series.timeframe, series.max_series_length)
        self.ma = smooth(self.tr, smoother, period)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, Bar bar, short new_item_started):
        if len(self.series) <= 1:
            return np.nan

        cdef double c1 = self.series[1].close
        cdef double h_l = abs(bar.high - bar.low)
        cdef double h_pc = abs(bar.high - c1)
        cdef double l_pc = abs(bar.low - c1)
        self.tr.update(time, max(h_l, h_pc, l_pc))
        return (100 * self.ma[0] / c1) if self.percentage else self.ma[0]


def atr(series: OHLCV, period: int = 14, smoother="sma", percentage: bool = False):
    if not isinstance(series, OHLCV):
        raise ValueError("Series must be OHLCV !")
    return Atr.wrap(series, period, smoother, percentage)


cdef class Zscore(Indicator):
    """
    Z-score normalization using rolling SMA and Std
    """

    def __init__(self, str name, TimeSeries series, int period, str smoother):
        self.tr = TimeSeries("tr", series.timeframe, series.max_series_length)
        self.ma = smooth(self.tr, smoother, period)
        self.std = std(self.tr, period)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        self.tr.update(time, value)
        if len(self.ma) < 1 or len(self.std) < 1 or np.isnan(self.ma[0]) or np.isnan(self.std[0]) or self.std[0] == 0:
            return np.nan
        return (value - self.ma[0]) / self.std[0]


def zscore(series: TimeSeries, period: int = 20, smoother="sma"):
    return Zscore.wrap(series, period, smoother)


cdef class BollingerBands(Indicator):
    """
    Bollinger Bands indicator using rolling SMA/other smoother and Std
    """

    def __init__(self, str name, TimeSeries series, int period, double nstd, str smoother):
        self.period = period
        self.nstd = nstd
        self.smoother = smoother
        
        # Create temporary series for tracking values
        self.tr = TimeSeries("tr", series.timeframe, series.max_series_length)
        
        # Create the moving average based on smoother type
        self.ma = smooth(self.tr, smoother, period)
        
        # Create standard deviation indicator with ddof=1 for sample std (matching pandas)
        self.std = std(self.tr, period, ddof=1)
        
        # Create upper and lower band series
        self.upper = TimeSeries("upper", series.timeframe, series.max_series_length)
        self.lower = TimeSeries("lower", series.timeframe, series.max_series_length)
        
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        # Update the tracking series
        self.tr.update(time, value)
        
        # Get the current moving average
        cdef double ma_value = self.ma[0] if len(self.ma) > 0 else np.nan
        
        # Get the current standard deviation
        cdef double std_value = self.std[0] if len(self.std) > 0 else np.nan
        
        # Declare band variables
        cdef double upper_band
        cdef double lower_band
        
        # Calculate bands if we have valid values
        if not np.isnan(ma_value) and not np.isnan(std_value):
            upper_band = ma_value + self.nstd * std_value
            lower_band = ma_value - self.nstd * std_value
            
            # Update the band series
            self.upper.update(time, upper_band)
            self.lower.update(time, lower_band)
            
            # Return the middle band (moving average)
            return ma_value
        else:
            # Update with NaN if we don't have enough data
            self.upper.update(time, np.nan)
            self.lower.update(time, np.nan)
            return np.nan


def bollinger_bands(series: TimeSeries, period: int = 20, nstd: float = 2.0, smoother: str = "sma"):
    """
    Bollinger Bands indicator
    
    :param series: Input time series
    :param period: Period for moving average and standard deviation (default 20)
    :param nstd: Number of standard deviations for bands (default 2.0)
    :param smoother: Type of smoother to use for middle band (default "sma")
    :return: BollingerBands indicator with middle, upper, and lower bands
    """
    return BollingerBands.wrap(series, period, nstd, smoother)


cdef class Swings(IndicatorOHLC):

    def __init__(self, str name, OHLCV series, trend_indicator, **indicator_args):
        self.base = OHLCV("base", series.timeframe, series.max_series_length)
        self.trend = trend_indicator(self.base, **indicator_args)

        self.tops = TimeSeries("tops", series.timeframe, series.max_series_length)
        self.tops_detection_lag = TimeSeries("tops_lag", series.timeframe, series.max_series_length)

        self.bottoms = TimeSeries("bottoms", series.timeframe, series.max_series_length)
        self.bottoms_detection_lag = TimeSeries("bottoms_lag", series.timeframe, series.max_series_length)

        self.middles = TimeSeries("middles", series.timeframe, series.max_series_length)
        self.deltas = TimeSeries("deltas", series.timeframe, series.max_series_length)

        # - store parameters for copying
        self._trend_indicator = trend_indicator
        self._indicator_args = indicator_args

        self._min_l = +np.inf
        self._max_h = -np.inf
        self._max_t = 0
        self._min_t = 0
        super().__init__(name, series)

    cpdef double calculate(self, long long time, Bar bar, short new_item_started):
        self.base.update_by_bar(time, bar.open, bar.high, bar.low, bar.close, bar.volume)
        cdef int _t = 0

        if len(self.trend.upper) > 0:
            _u = self.trend.upper[0]
            _d = self.trend.lower[0]

            if not np.isnan(_u):
                if self._max_t > 0:
                    self.tops.update(self._max_t, self._max_h)
                    self.tops_detection_lag.update(self._max_t, time - self._max_t)
                    if len(self.bottoms) > 0:
                        self.middles.update(time, (self.tops[0] + self.bottoms[0]) / 2)
                        self.deltas.update(time, self.tops[0] - self.bottoms[0])

                if bar.low <= self._min_l:
                    self._min_l = bar.low
                    self._min_t = time

                self._max_h = -np.inf
                self._max_t = 0
                _t = -1
            elif not np.isnan(_d):
                if self._min_t > 0:
                    self.bottoms.update(self._min_t, self._min_l)
                    self.bottoms_detection_lag.update(self._min_t, time - self._min_t)
                    if len(self.tops) > 0:
                        self.middles.update(time, (self.tops[0] + self.bottoms[0]) / 2)
                        self.deltas.update(time, self.tops[0] - self.bottoms[0])

                if bar.high >= self._max_h:
                    self._max_h = bar.high
                    self._max_t = time

                self._min_l = +np.inf
                self._min_t = 0
                _t = +1

        return _t

    def get_current_trend_end(self):
        if np.isfinite(self._min_l):
            return pd.Timestamp(self._min_t, 'ns'), self._min_l
        elif np.isfinite(self._max_h):
            return pd.Timestamp(self._max_t, 'ns'), self._max_h
        return (None, None)

    def copy(self, int start, int stop):
        n_ts = Swings(self.name, OHLCV("base", self.series.timeframe), self._trend_indicator, **self._indicator_args)

        # - copy main series
        for i in range(start, stop):
            n_ts._add_new_item(self.times.values[i], self.values.values[i])
            n_ts.trend._add_new_item(self.trend.times.values[i], self.trend.values.values[i])

        # - copy internal series
        (
            n_ts.tops, 
            n_ts.tops_detection_lag,
            n_ts.bottoms,
            n_ts.bottoms_detection_lag,
            n_ts.middles,
            n_ts.deltas
        ) = self._copy_internal_series(start, stop, 
            self.tops, 
            self.tops_detection_lag,
            self.bottoms,
            self.bottoms_detection_lag,
            self.middles,
            self.deltas
        )

        return n_ts

    def pd(self) -> pd.DataFrame:
        _t, _d = self.get_current_trend_end()
        tps, bts = self.tops.pd(), self.bottoms.pd()
        tpl, btl = self.tops_detection_lag.pd(), self.bottoms_detection_lag.pd()
        if _t is not None:
            if bts.index[-1] < tps.index[-1]:
                bts = srows(bts, pd.Series({_t: _d}))
                btl = srows(btl, pd.Series({_t: 0}))  # last lag is 0
            else:
                tps = srows(tps, pd.Series({_t: _d}))
                tpl = srows(tpl, pd.Series({_t: 0})) # last lag is 0

        # - convert tpl / btl to timedeltas
        tpl, btl = tpl.apply(lambda x: pd.Timedelta(x, unit='ns')), btl.apply(lambda x: pd.Timedelta(x, unit='ns'))

        eid = pd.Series(tps.index, tps.index)
        mx = scols(bts, tps, eid, names=["start_price", "end_price", "end"])
        dt = scols(mx["start_price"], mx["end_price"].shift(-1), mx["end"].shift(-1))  # .dropna()
        dt = dt.assign(
            delta = dt["end_price"] - dt["start_price"], 
            spotted = pd.Series(bts.index + btl, bts.index)
        )

        eid = pd.Series(bts.index, bts.index)
        mx = scols(tps, bts, eid, names=["start_price", "end_price", "end"])
        ut = scols(mx["start_price"], mx["end_price"].shift(-1), mx["end"].shift(-1))  # .dropna()
        ut = ut.assign(
            delta = ut["end_price"] - ut["start_price"], 
            spotted = pd.Series(tps.index + tpl, tps.index)
        )

        return scols(ut, dt, keys=["DownTrends", "UpTrends"])


def swings(series: OHLCV, trend_indicator, **indicator_args):
    """
    Swing detector based on provided trend indicator.
    """
    if not isinstance(series, OHLCV):
        raise ValueError("Series must be OHLCV !")
    return Swings.wrap(series, trend_indicator, **indicator_args)


cdef class Pivots(IndicatorOHLC):
    """
    Pivot points detector that identifies local highs and lows using
    lookback (before) and lookahead (after) windows.
    """

    def __init__(self, str name, OHLCV series, int before, int after):
        self.before = before
        self.after = after
        
        # Deque to store completed bars for pivot detection
        self.bars_buffer = deque(maxlen=before + after + 1)
        
        # Keep track of the current unfinished bar separately
        self.current_bar = None
        self.current_bar_time = 0
        
        # TimeSeries for pivot points
        self.tops = TimeSeries("tops", series.timeframe, series.max_series_length)
        self.bottoms = TimeSeries("bottoms", series.timeframe, series.max_series_length)
        self.tops_detection_lag = TimeSeries("tops_lag", series.timeframe, series.max_series_length)
        self.bottoms_detection_lag = TimeSeries("bottoms_lag", series.timeframe, series.max_series_length)
        
        super().__init__(name, series)
    
    cpdef double calculate(self, long long time, Bar bar, short new_item_started):
        cdef int pivot_idx, i
        cdef double pivot_high, pivot_low
        cdef long long pivot_time
        cdef short is_pivot_high, is_pivot_low
        
        if new_item_started:
            # If we have a previous bar that was being updated, add it to the buffer as completed
            if self.current_bar is not None and self.current_bar_time > 0:
                self.bars_buffer.append((self.current_bar_time, self.current_bar))
            
            # Start tracking the new unfinished bar
            self.current_bar = bar
            self.current_bar_time = time
            
            # Check if we have enough completed bars to detect a pivot
            # We need exactly before + after + 1 bars in the buffer
            if len(self.bars_buffer) < self.before + self.after + 1:
                return np.nan
            
            # The pivot candidate is at index 'before'
            pivot_idx = self.before
            pivot_time = self.bars_buffer[pivot_idx][0]
            pivot_bar = self.bars_buffer[pivot_idx][1]
            pivot_high = pivot_bar.high
            pivot_low = pivot_bar.low
            
            # Check for pivot high: pivot high must be > all other highs in window
            is_pivot_high = 1
            for i in range(len(self.bars_buffer)):
                if i != pivot_idx:
                    if self.bars_buffer[i][1].high >= pivot_high:
                        is_pivot_high = 0
                        break
            
            # Check for pivot low: pivot low must be < all other lows in window
            is_pivot_low = 1
            for i in range(len(self.bars_buffer)):
                if i != pivot_idx:
                    if self.bars_buffer[i][1].low <= pivot_low:
                        is_pivot_low = 0
                        break
            
            # Record pivot high if found
            if is_pivot_high:
                self.tops.update(pivot_time, pivot_high)
                # Detection time is now (when we actually detect it)
                self.tops_detection_lag.update(pivot_time, time - pivot_time)
            
            # Record pivot low if found
            if is_pivot_low:
                self.bottoms.update(pivot_time, pivot_low)
                # Detection time is now (when we actually detect it)
                self.bottoms_detection_lag.update(pivot_time, time - pivot_time)
            
            # Return 1 for pivot high, -1 for pivot low, 0 for both, nan for neither
            if is_pivot_high and is_pivot_low:
                return 0
            elif is_pivot_high:
                return 1
            elif is_pivot_low:
                return -1
            else:
                return np.nan
        else:
            # Just update the current unfinished bar
            self.current_bar = bar
            # Note: current_bar_time stays the same since we're updating the same bar
            return np.nan

    def pd(self) -> pd.DataFrame:
        """
        Return DataFrame with pivot points and detection lags.
        
        Returns a multi-column DataFrame with:
        - Tops: price, detection_lag, spotted (time when pivot was detected)
        - Bottoms: price, detection_lag, spotted
        """
        from qubx.pandaz.utils import scols
        
        tps = self.tops.pd()
        bts = self.bottoms.pd()
        tpl = self.tops_detection_lag.pd()
        btl = self.bottoms_detection_lag.pd()
        
        # Convert lags to timedeltas
        if len(tpl) > 0:
            tpl = tpl.apply(lambda x: pd.Timedelta(x, unit='ns'))
        if len(btl) > 0:
            btl = btl.apply(lambda x: pd.Timedelta(x, unit='ns'))
        
        # Create DataFrames for tops and bottoms
        if len(tps) > 0:
            tops_df = pd.DataFrame({
                'price': tps,
                'detection_lag': tpl,
                'spotted': pd.Series(tps.index + tpl.values, index=tps.index)
            })
        else:
            tops_df = pd.DataFrame(columns=['price', 'detection_lag', 'spotted'])
        
        if len(bts) > 0:
            bottoms_df = pd.DataFrame({
                'price': bts,
                'detection_lag': btl,
                'spotted': pd.Series(bts.index + btl.values, index=bts.index)
            })
        else:
            bottoms_df = pd.DataFrame(columns=['price', 'detection_lag', 'spotted'])
        
        return scols(tops_df, bottoms_df, keys=["Tops", "Bottoms"])


def pivots(series: OHLCV, before: int = 5, after: int = 5):
    """
    Pivot points detector using lookback/lookahead windows.

    :param series: OHLCV series
    :param before: Number of bars to look back
    :param after: Number of bars to look ahead
    :return: Pivots indicator with tops and bottoms
    """
    if not isinstance(series, OHLCV):
        raise ValueError("Series must be OHLCV!")
    return Pivots.wrap(series, before, after)


cdef class PctChange(Indicator):
    """
    Percentage change indicator that calculates the percentage change
    between current value and value from n periods ago.

    Note: Returns percentage as a decimal (0.01 for 1%), matching pandas behavior.
    """
    def __init__(self, str name, TimeSeries series, int period):
        self.period = period
        if period <= 0:
            raise ValueError("Period must be positive and greater than zero")

        # Buffer to store past values for the specified period
        self.past_values = deque(maxlen=period + 1)
        self._count = 0

        # - store/restore pattern for handling bar updates
        self.stored_past_values = None
        self.stored_count = 0

        super().__init__(name, series)

    cdef void _store(self):
        """Store current state before bar update"""
        self.stored_past_values = deque(self.past_values, maxlen=self.period + 1)
        self.stored_count = self._count

    cdef void _restore(self):
        """Restore state for bar update"""
        if self.stored_past_values is not None:
            self.past_values = deque(self.stored_past_values, maxlen=self.period + 1)
            self._count = self.stored_count

    cpdef double calculate(self, long long time, double value, short new_item_started):
        cdef double prev_value
        cdef double result

        # - If this is the very first value (indicator created on empty series),
        # - treat it as a new item regardless of new_item_started flag
        if len(self.past_values) == 0:
            new_item_started = True

        # - store/restore pattern for bar updates
        if not new_item_started:
            self._restore()

        if new_item_started:
            # New bar started, add value to history
            self.past_values.append(value)
            self._count += 1
            # Store state for potential bar updates
            self._store()
        else:
            # Updating existing bar - update the last value
            if len(self.past_values) > 0:
                self.past_values[-1] = value

        # Calculate percentage change if we have enough history
        if len(self.past_values) > self.period:
            # Get the value from 'period' bars ago
            prev_value = self.past_values[-(self.period + 1)]

            # Handle zero or NaN previous value
            if np.isnan(prev_value) or prev_value == 0:
                return np.nan

            # Calculate percentage change (as decimal, like pandas)
            result = (value - prev_value) / prev_value
            return result
        else:
            # Not enough history yet
            return np.nan


def pct_change(series: TimeSeries, period: int = 1):
    """
    Calculate percentage change between current value and value from n periods ago.

    Returns the percentage change as a decimal (e.g., 0.01 for 1% increase),
    matching the behavior of pandas.DataFrame.pct_change().

    :param series: Input time series
    :param period: Number of periods to shift for calculating percentage change (default 1)
    :return: PctChange indicator
    """
    return PctChange.wrap(series, period)


cdef class Rsi(Indicator):
    """
    Relative Strength Index indicator

    RSI measures the magnitude of recent price changes to evaluate
    overbought or oversold conditions.

    Formula:
        U = max(0, price_change)
        D = max(0, -price_change)
        RS = smooth(U) / smooth(D)
        RSI = 100 - (100 / (1 + RS))

    Or equivalently:
        RSI = 100 * smooth(U) / (smooth(U) + smooth(D))
    """

    def __init__(self, str name, TimeSeries series, int period, str smoother="ema"):
        self.period = period
        self.prev_value = np.nan

        # - create internal series for up and down moves
        self.up_moves = TimeSeries("up_moves", series.timeframe, series.max_series_length)
        self.down_moves = TimeSeries("down_moves", series.timeframe, series.max_series_length)

        # - create smoothers for up and down moves
        self.smooth_up = smooth(self.up_moves, smoother, period)
        self.smooth_down = smooth(self.down_moves, smoother, period)

        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        cdef double diff, up_move, down_move, smooth_u, smooth_d

        # - for the first value, store it and return NaN
        if np.isnan(self.prev_value):
            self.prev_value = value
            self.up_moves.update(time, 0.0)
            self.down_moves.update(time, 0.0)
            return np.nan

        # - calculate price change
        if new_item_started:
            diff = value - self.prev_value
            self.prev_value = value
        else:
            # - when updating the same bar, calculate diff from previous closed bar
            diff = value - self.prev_value

        # - calculate up and down moves
        if diff > 0:
            up_move = diff
            down_move = 0.0
        elif diff < 0:
            up_move = 0.0
            down_move = -diff  # - make it positive
        else:
            up_move = 0.0
            down_move = 0.0

        # - update internal series
        self.up_moves.update(time, up_move)
        self.down_moves.update(time, down_move)

        # - get smoothed values
        smooth_u = self.smooth_up[0]
        smooth_d = self.smooth_down[0]

        # - check if we have valid smoothed values
        if np.isnan(smooth_u) or np.isnan(smooth_d):
            return np.nan

        # - avoid division by zero
        if smooth_u + smooth_d == 0:
            return 50.0  # - neutral value when there's no movement

        # - calculate RSI
        return 100.0 * smooth_u / (smooth_u + smooth_d)


def rsi(series: TimeSeries, period: int = 14, smoother: str = "ema"):
    """
    Relative Strength Index indicator

    The RSI is a momentum oscillator that measures the speed and magnitude
    of recent price changes to evaluate overbought or oversold conditions.

    :param series: Input time series
    :param period: Number of periods for smoothing (default 14)
    :param smoother: Smoothing method: 'sma', 'ema', 'tema', 'dema', 'kama' (default 'ema')
    :return: RSI indicator (values range from 0 to 100)
    """
    return Rsi.wrap(series, period, smoother) # type: ignore


cdef class StdEma(Indicator):
    """
    Standard deviation using exponential weighted moving average.

    Calculates the EWM standard deviation of input series values.
    Works on any TimeSeries (returns, price differences, etc.).
    Matches pandas.Series.ewm(span=period).std() behavior (default adjust=True).

    Uses incremental algorithm for O(1) updates instead of O(n).
    """

    def __init__(self, str name, TimeSeries series, int period):
        self.period = period
        self.alpha = 2.0 / (period + 1.0)
        self.count = 0

        # - incremental EWM accumulators (for adjust=True mode)
        self.ewm_mean_numer = 0.0  # - sum(w_i * x_i)
        self.ewm_mean_denom = 0.0  # - sum(w_i)
        self.ewm_var_numer = 0.0   # - sum(w_i * (x_i - mean)^2)
        self.ewm_var_denom = 0.0   # - sum(w_i) for variance
        self.prev_mean = 0.0

        # - store/restore pattern for handling bar updates
        self.stored_count = 0
        self.stored_ewm_mean_numer = 0.0
        self.stored_ewm_mean_denom = 0.0
        self.stored_ewm_var_numer = 0.0
        self.stored_ewm_var_denom = 0.0
        self.stored_prev_mean = 0.0

        super().__init__(name, series)

    cdef void _store(self):
        """Store current state before bar update"""
        self.stored_count = self.count
        self.stored_ewm_mean_numer = self.ewm_mean_numer
        self.stored_ewm_mean_denom = self.ewm_mean_denom
        self.stored_ewm_var_numer = self.ewm_var_numer
        self.stored_ewm_var_denom = self.ewm_var_denom
        self.stored_prev_mean = self.prev_mean

    cdef void _restore(self):
        """Restore state for bar update (don't restore count)"""
        # - Note: We don't restore count because it should only increment on new bars
        self.ewm_mean_numer = self.stored_ewm_mean_numer
        self.ewm_mean_denom = self.stored_ewm_mean_denom
        self.ewm_var_numer = self.stored_ewm_var_numer
        self.ewm_var_denom = self.stored_ewm_var_denom
        self.prev_mean = self.stored_prev_mean

    cpdef double calculate(self, long long time, double value, short new_item_started):
        cdef double w_decay, new_weight, delta, delta_new
        cdef double current_mean, old_value_estimate, numer_without_last, var_without_last

        # - skip NaN values
        if np.isnan(value):
            return np.nan

        # - If this is the very first value (indicator created on empty series),
        # - treat it as a new item regardless of new_item_started flag
        if self.count == 0:
            new_item_started = True

        # - store/restore pattern for bar updates
        if not new_item_started:
            self._restore()

        # - weight decay factor: all existing weights get multiplied by (1-alpha)
        w_decay = 1.0 - self.alpha

        # - new observation gets weight 1.0
        new_weight = 1.0

        if new_item_started:
            # Store state BEFORE adding new bar
            self._store()

            # - decay all previous weights
            self.ewm_mean_numer = w_decay * self.ewm_mean_numer + new_weight * value
            self.ewm_mean_denom = w_decay * self.ewm_mean_denom + new_weight

            # - calculate current mean
            current_mean = self.ewm_mean_numer / self.ewm_mean_denom if self.ewm_mean_denom > 0 else 0.0

            # - for variance, we need to update with the new delta
            # - when adding a new point, we update variance incrementally
            delta = value - self.prev_mean
            delta_new = value - current_mean

            self.ewm_var_numer = w_decay * self.ewm_var_numer + new_weight * delta * delta_new
            self.ewm_var_denom = w_decay * self.ewm_var_denom + new_weight

            self.prev_mean = current_mean
            self.count += 1
        else:
            # - bar update: after restore, we're back to state BEFORE current bar
            # - apply same logic as new item (decay and add), but don't increment count
            self.ewm_mean_numer = w_decay * self.ewm_mean_numer + new_weight * value
            self.ewm_mean_denom = w_decay * self.ewm_mean_denom + new_weight

            # - calculate current mean
            current_mean = self.ewm_mean_numer / self.ewm_mean_denom if self.ewm_mean_denom > 0 else 0.0

            # - update variance
            delta = value - self.prev_mean
            delta_new = value - current_mean

            self.ewm_var_numer = w_decay * self.ewm_var_numer + new_weight * delta * delta_new
            self.ewm_var_denom = w_decay * self.ewm_var_denom + new_weight

            self.prev_mean = current_mean

        # - need at least `period` values before we can calculate
        if self.count < self.period:
            return np.nan

        # - calculate EWM variance with bias correction
        # - for adjust=True mode, we need sum(w_i^2) / sum(w_i)
        # - for weights w_i = (1-alpha)^(n-1-i), we can compute this incrementally
        cdef double sum_weights = self.ewm_var_denom
        cdef double r_sq = w_decay * w_decay
        cdef double sum_weights_sq
        cdef int n
        cdef double bias_correction
        cdef double ewm_var

        # - sum of squared weights: sum((1-alpha)^(2*(n-1-i)))
        # - this is a geometric series: 1 + r^2 + r^4 + ... where r = (1-alpha)
        # - for large n, this converges to 1 / (1 - r^2) = 1 / (1 - (1-alpha)^2) = 1 / (alpha * (2-alpha))
        # - but we need exact value for finite n

        # - use the asymptotic approximation (valid for n >> period)
        if self.count > self.period * 2:
            # - use asymptotic formula for efficiency
            sum_weights_sq = 1.0 / (1.0 - r_sq) if r_sq < 1.0 else sum_weights
        else:
            # - for early values, compute exactly
            # - sum_sq = sum((1-alpha)^(2i)) for i=0 to n-1 = (1 - r^(2n)) / (1 - r^2)
            n = self.count
            sum_weights_sq = (1.0 - r_sq ** n) / (1.0 - r_sq) if r_sq < 1.0 else <double>n

        # - bias correction factor
        bias_correction = sum_weights - sum_weights_sq / sum_weights if sum_weights > 0 else 1.0
        ewm_var = self.ewm_var_numer / bias_correction if bias_correction > 0 else 0.0

        # - handle numerical errors
        if ewm_var < 0:
            ewm_var = 0.0

        # - return standard deviation
        return np.sqrt(ewm_var)


def stdema(series: TimeSeries, period: int):
    """
    Calculate exponential weighted moving standard deviation. This is equivalent of 
    ```
    series.ewm(span=period, min_periods=period).std()
    ```

    Parameters
    ----------
    series : TimeSeries
        Input series (returns, price differences, etc.)
    period : int
        EWM span parameter

    Returns
    -------
    StdEma
        Standard deviation indicator
    """
    return StdEma.wrap(series, period) # type: ignore


cdef class CusumFilter(Indicator):
    """
    CUSUM filter on streaming data

    Detects significant changes in a time series by tracking cumulative deviations.
    Returns 1 when an event is detected (cumulative change exceeds threshold), 0 otherwise.

    The algorithm tracks both positive and negative cumulative sums and triggers events
    when either exceeds the threshold. The threshold is dynamically calculated from a
    target series (e.g., volatility) multiplied by the current value.
    """

    def __init__(self, str name, TimeSeries series, TimeSeries target):
        self.target_cache = SeriesCachedValue(target)
        self.s_pos = 0.0
        self.s_neg = 0.0
        self.prev_value = np.nan
        self.last_value = np.nan  # - Track last processed value
        self.prev_bar_event = 0.0  # - Event from previous completed bar (what we return)
        self.current_bar_event = 0.0  # - Event for current bar
        super().__init__(name, series)

    cdef void _store(self):
        """
        Store state before processing update
        """
        self.saved_s_pos = self.s_pos
        self.saved_s_neg = self.s_neg
        self.saved_prev_value = self.prev_value

    cdef void _restore(self):
        """
        Restore state when bar is updated (not new)
        Restores cumulative sums but NOT prev_value
        prev_value must remain as previous bar's final value for correct diff calculation
        """
        self.s_pos = self.saved_s_pos
        self.s_neg = self.saved_s_neg
        # - DON'T restore prev_value - keep it as previous bar's final value

    cpdef double calculate(self, long long time, double value, short new_item_started):
        cdef double diff, threshold, target_value
        cdef int event = 0
        cdef double return_value

        # - first value - just store it
        if np.isnan(self.prev_value):
            self.prev_value = value
            self.last_value = value
            self._store()
            self.current_bar_event = 0.0
            return 0.0

        # - for new bar, update prev_value to last bar's final value
        if new_item_started:
            self.prev_value = self.last_value

        # - for intrabar updates, restore state
        if not new_item_started:
            self._restore()
        else:
            # - for new bar: check if this is OPEN tick (value ~ last_value) or direct CLOSE
            # - if OPEN tick, skip processing and just store state for intrabar updates
            # - if direct CLOSE (static OHLC), process normally
            # - use relative threshold (0.01% of price) to handle different price ranges
            threshold = abs(self.last_value * 0.0001) + 0.001  # 0.01% + small absolute
            if abs(value - self.last_value) < threshold:  # OPEN tick (value ~ previous CLOSE)
                # - store current state without processing
                self._store()
                self.last_value = value
                self.prev_bar_event = self.current_bar_event
                self.current_bar_event = 0.0
                return self.prev_bar_event

        # - calculate diff
        diff = value - self.prev_value

        # - update cumulative sums (do this regardless of threshold availability)
        self.s_pos = max(0.0, self.s_pos + diff)
        self.s_neg = min(0.0, self.s_neg + diff)

        # - get threshold from target series (it can be from higher timeframe)
        # - to get the FINAL value from previous period (not partial value from same time yesterday),
        # - we need to look up just before the start of current period
        # - this ensures we use yesterday's FINAL volatility for today's CUSUM calculations
        cdef long long current_period_start = floor_t64(time, self.target_cache.ser.timeframe)
        cdef long long lookup_time = current_period_start - 1  # - 1 nanosecond before current period
        target_value = self.target_cache.value(lookup_time)

        # - check for events if threshold is available
        if not np.isnan(target_value):
            threshold = abs(target_value * value)

            # - check for events (comparisons with valid threshold)
            if self.s_neg < -threshold:
                self.s_neg = 0.0
                event = 1
            elif self.s_pos > threshold:
                self.s_pos = 0.0
                event = 1

        # - Store state and return event
        # - Both static OHLC and tick-based CLOSE should return current event
        if new_item_started:
            # - new bar starting (static OHLC with CLOSE value, or tick-based OPEN - but OPEN was skipped above)
            self.last_value = value
            self._store()
            return float(event)
        else:
            # - intrabar update (tick-based: HIGH, LOW, or CLOSE ticks)
            self.last_value = value
            # - DO NOT store: all intrabar updates restore to bar start
            # - return current event (CLOSE tick event)
            return float(event)


def cusum_filter(series: TimeSeries, target: TimeSeries):
    """
    Cusum filter implementation for streaming data.

    Detects significant changes in a time series by tracking cumulative deviations.
    Returns a TimeSeries with 1 at event points (when cumulative change exceeds threshold)
    and 0 elsewhere.

    Parameters
    ----------
    series : TimeSeries
        Input series (price series, etc.)
    target : TimeSeries
        Target threshold series (typically volatility, can be from higher timeframe like daily)
        The threshold is calculated as target * current_price

    Returns
    -------
    CusumFilter
        CusumFilter indicator that returns 1 when event is detected, 0 otherwise

    Examples
    --------
    >>> # - daily volatility
    >>> vol = stdema(pct_change(daily_close), 30)
    >>> # - detect significant moves on hourly data using daily volatility
    >>> events = cusum_filter(hourly_close, vol * 2)
    """
    return CusumFilter.wrap(series, target) # type: ignore


cdef class Macd(Indicator):
    """
    Moving Average Convergence Divergence (MACD) indicator

    MACD is calculated as:
    1. MACD Line = fast_ma(price) - slow_ma(price)
    2. Signal Line = signal_ma(MACD Line)

    The returned value is the Signal Line (the smoothed MACD).
    """
    def __init__(self, str name, TimeSeries series, fast=12, slow=26, signal=9, method="ema", signal_method="ema"):
        self.fast_period = fast
        self.slow_period = slow
        self.signal_period = signal
        self.method = method
        self.signal_method = signal_method

        # - create internal copy of input series to track values
        self.input_series = TimeSeries("input", series.timeframe, series.max_series_length)

        # - create fast and slow moving averages on the internal series
        self.fast_ma = smooth(self.input_series, method, fast)
        self.slow_ma = smooth(self.input_series, method, slow)

        # - create internal series for MACD line (fast - slow)
        self.macd_line_series = TimeSeries("macd_line", series.timeframe, series.max_series_length)

        # - create signal line (smoothed MACD line)
        self.signal_line = smooth(self.macd_line_series, signal_method, signal)

        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        cdef double fast_value, slow_value, macd_value

        # - update internal input series first
        self.input_series.update(time, value)

        # - get fast and slow MA values (they are automatically calculated)
        fast_value = self.fast_ma[0] if len(self.fast_ma) > 0 else np.nan
        slow_value = self.slow_ma[0] if len(self.slow_ma) > 0 else np.nan

        # - calculate MACD line (fast - slow)
        if np.isnan(fast_value) or np.isnan(slow_value):
            macd_value = np.nan
        else:
            macd_value = fast_value - slow_value

        # - update MACD line series
        self.macd_line_series.update(time, macd_value)

        # - return signal line value (smoothed MACD)
        return self.signal_line[0] if len(self.signal_line) > 0 else np.nan


def macd(series: TimeSeries, fast=12, slow=26, signal=9, method="ema", signal_method="ema"):
    """
    Moving average convergence divergence (MACD) is a trend-following momentum indicator that shows the relationship
    between two moving averages of prices. The MACD is calculated by subtracting the 26-day slow moving average from the
    12-day fast MA. A nine-day MA of the MACD, called the "signal line", is then plotted on top of the MACD,
    functioning as a trigger for buy and sell signals.

    :param series: input data
    :param fast: fast MA period
    :param slow: slow MA period
    :param signal: signal MA period
    :param method: used moving averaging method (sma, ema, tema, dema, kama)
    :param signal_method: used method for averaging signal (sma, ema, tema, dema, kama)
    :return: macd indicator
    """
    return Macd.wrap(series, fast, slow, signal, method, signal_method) # type: ignore


cdef class SuperTrend(IndicatorOHLC):
    """
    SuperTrend indicator - a trend-following indicator based on ATR

    The SuperTrend indicator provides trend direction and support/resistance levels.
    It uses Average True Range (ATR) to calculate dynamic bands around price.

    Returns:
        trend: +1 for uptrend, -1 for downtrend
        utl (upper trend line): longstop values during uptrend
        dtl (down trend line): shortstop values during downtrend
    """

    def __init__(self, str name, OHLCV series, int length, double mult, str src, short wicks, str atr_smoother):
        self.length = length
        self.mult = mult
        self.src = src
        self.wicks = wicks
        self.atr_smoother = atr_smoother

        # - working state variables (updated during calculation)
        self._prev_longstop = np.nan
        self._prev_shortstop = np.nan
        self._prev_direction = np.nan

        # - saved state variables (for handling partial bar updates)
        self.prev_longstop = np.nan
        self.prev_shortstop = np.nan
        self.prev_direction = np.nan

        # - create internal TR series and smooth it for ATR calculation
        self.tr = TimeSeries("tr", series.timeframe, series.max_series_length)
        self.atr_ma = smooth(self.tr, atr_smoother, length)

        # - create output series for upper and down trend lines
        self.utl = TimeSeries("utl", series.timeframe, series.max_series_length)
        self.dtl = TimeSeries("dtl", series.timeframe, series.max_series_length)

        super().__init__(name, series)

    cdef _store(self):
        """Store current working state to saved state"""
        self.prev_longstop = self._prev_longstop
        self.prev_shortstop = self._prev_shortstop
        self.prev_direction = self._prev_direction

    cdef _restore(self):
        """Restore saved state to working state"""
        self._prev_longstop = self.prev_longstop
        self._prev_shortstop = self.prev_shortstop
        self._prev_direction = self.prev_direction

    cdef double calc_src(self, Bar bar):
        """Calculate source value based on src parameter"""
        if self.src == "close":
            return bar.close
        elif self.src == "hl2":
            return (bar.high + bar.low) / 2.0
        elif self.src == "hlc3":
            return (bar.high + bar.low + bar.close) / 3.0
        elif self.src == "ohlc4":
            return (bar.open + bar.high + bar.low + bar.close) / 4.0
        else:
            return (bar.high + bar.low) / 2.0  # - default to hl2

    cpdef double calculate(self, long long time, Bar bar, short new_item_started):
        cdef double atr_value, src_value, longstop, shortstop
        cdef double high_price, low_price, p_high_price, p_low_price
        cdef short is_doji4price
        cdef double direction
        cdef double saved_prev_longstop, saved_prev_shortstop
        cdef double tr_value

        # - need at least 2 bars for prev calculations
        if len(self.series) < 2:
            # - initialize on first bar
            self._prev_longstop = np.nan
            self._prev_shortstop = np.nan
            self._prev_direction = np.nan
            self._store()
            return np.nan

        # - handle store/restore for partial bar updates
        if new_item_started:
            self._store()
        else:
            self._restore()

        # - calculate True Range
        cdef Bar prev_bar = self.series[1]
        cdef double c1 = prev_bar.close
        cdef double h_l = abs(bar.high - bar.low)
        cdef double h_pc = abs(bar.high - c1)
        cdef double l_pc = abs(bar.low - c1)
        tr_value = max(h_l, h_pc, l_pc)

        # - update TR series (this will automatically update ATR via smooth)
        self.tr.update(time, tr_value)

        # - get ATR value
        atr_value = self.atr_ma[0]
        if np.isnan(atr_value):
            return np.nan

        atr_value = abs(self.mult) * atr_value

        # - calculate source value
        src_value = self.calc_src(bar)

        # - determine which prices to use for wicks
        if self.wicks:
            high_price = bar.high
            low_price = bar.low
        else:
            high_price = bar.close
            low_price = bar.close

        # - get previous bar's prices (prev_bar already retrieved above for TR)
        if self.wicks:
            p_high_price = prev_bar.high
            p_low_price = prev_bar.low
        else:
            p_high_price = prev_bar.close
            p_low_price = prev_bar.close

        # - check for doji4price (all prices equal)
        is_doji4price = (bar.open == bar.close) and (bar.open == bar.low) and (bar.open == bar.high)

        # - calculate basic stops
        longstop = src_value - atr_value
        shortstop = src_value + atr_value

        # - save previous bar's stops for direction comparison
        saved_prev_longstop = self._prev_longstop
        saved_prev_shortstop = self._prev_shortstop

        # - adjust longstop based on previous value
        if np.isnan(self._prev_longstop):
            self._prev_longstop = longstop

        if longstop > 0:
            if is_doji4price:
                longstop = self._prev_longstop
            else:
                if p_low_price > self._prev_longstop:
                    longstop = max(longstop, self._prev_longstop)
                # - else: keep calculated longstop
        else:
            longstop = self._prev_longstop

        # - adjust shortstop based on previous value
        if np.isnan(self._prev_shortstop):
            self._prev_shortstop = shortstop

        if shortstop > 0:
            if is_doji4price:
                shortstop = self._prev_shortstop
            else:
                if p_high_price < self._prev_shortstop:
                    shortstop = min(shortstop, self._prev_shortstop)
                # - else: keep calculated shortstop
        else:
            shortstop = self._prev_shortstop

        # - determine direction based on price breaking PREVIOUS bar's stops
        # - only check for breaks if we have valid previous stops
        if np.isnan(saved_prev_longstop) or np.isnan(saved_prev_shortstop):
            # - not enough data yet, forward fill previous direction or return NaN
            direction = self._prev_direction if not np.isnan(self._prev_direction) else np.nan
        elif low_price < saved_prev_longstop:
            direction = -1.0
        elif high_price > saved_prev_shortstop:
            direction = 1.0
        else:
            # - no break, keep previous direction (forward fill)
            direction = self._prev_direction if not np.isnan(self._prev_direction) else np.nan

        # - update working state for next iteration
        self._prev_longstop = longstop
        self._prev_shortstop = shortstop
        self._prev_direction = direction

        # - update utl and dtl series based on direction
        # - only update when we have a valid direction
        if direction == 1.0:
            self.utl.update(time, longstop)
        elif direction == -1.0:
            self.dtl.update(time, shortstop)

        # - return trend direction
        return direction


def super_trend(series: OHLCV, length: int = 22, mult: float = 3.0, src: str = "hl2",
                wicks: bool = True, atr_smoother: str = "sma"):
    """
    SuperTrend indicator - a trend-following indicator based on ATR

    The SuperTrend indicator uses Average True Range (ATR) to calculate dynamic support
    and resistance levels. It provides clear trend direction and can be used for
    entry/exit signals.

    :param series: OHLCV input series
    :param length: ATR period (default 22)
    :param mult: ATR multiplier (default 3.0)
    :param src: source calculation - 'close', 'hl2', 'hlc3', 'ohlc4' (default 'hl2')
    :param wicks: whether to use high/low (True) or close (False) for calculations (default True)
    :param atr_smoother: smoothing method for ATR - 'sma', 'ema', etc (default 'sma')
    :return: SuperTrend indicator with trend, utl, and dtl series
    """
    if not isinstance(series, OHLCV):
        raise ValueError("Series must be OHLCV !")
    return SuperTrend.wrap(series, length, mult, src, wicks, atr_smoother) # type: ignore


"""
Lighter exchange-specific API extensions.

Provides access to Lighter-specific features like leverage management,
transfers, and liquidity pool creation.
"""

# TODO: implement this

from typing import TYPE_CHECKING

from qubx import logger
from qubx.core.basics import Instrument
from qubx.core.interfaces import IExchangeExtensions

if TYPE_CHECKING:
    from .broker import LighterBroker
    from .client import LighterClient


class LighterExchangeAPI(IExchangeExtensions):
    """
    Exchange-specific API for Lighter.

    Provides access to Lighter-specific operations:
    - update_leverage: Update leverage for an instrument
    - transfer: Transfer USDC between accounts
    - create_pool: Create a public liquidity pool
    - create_subaccount: Create a new sub-account
    """

    def __init__(self, client: "LighterClient", broker: "LighterBroker"):
        """
        Initialize Lighter extensions.

        Args:
            client: LighterClient instance for API access
            broker: LighterBroker instance for instrument lookups
        """
        self.client = client
        self.broker = broker

    async def update_leverage(
        self, instrument: Instrument, leverage: float, margin_mode: str = "cross"
    ) -> tuple[bool, str | None]:
        """
        Update leverage for an instrument.

        Args:
            instrument: Instrument to update leverage for
            leverage: Target leverage ratio (e.g., 10.0 for 10x)
            margin_mode: "cross" or "isolated" (default: "cross")

        Returns:
            Tuple of (success: bool, error: str | None)

        Example:
            >>> # Update leverage to 10x for BTC/USDC
            >>> broker = ctx.get_broker("LIGHTER")
            >>> btc_usdc = ctx.query_instrument("BTC/USDC", "LIGHTER")
            >>> success, error = await broker.extensions.update_leverage(btc_usdc, 10.0)
            >>> if success:
            >>>     print("Leverage updated successfully")
            >>> else:
            >>>     print(f"Failed to update leverage: {error}")
        """
        try:
            # Get market_id from instrument via broker's instrument loader
            market_id = self.broker.instrument_loader.get_market_id(instrument.symbol)
            if market_id is None:
                return False, f"Instrument {instrument.symbol} has no market_id"

            # Convert margin mode to integer (0=cross, 1=isolated)
            margin_mode_int = 0 if margin_mode.lower() == "cross" else 1

            logger.info(
                f"Updating leverage for {instrument.symbol} (market_id={market_id}): "
                f"leverage={leverage}x, margin_mode={margin_mode}"
            )

            # Use SignerClient to update leverage
            # The update_leverage method takes: market_index, margin_mode, leverage
            result = await self.client.signer_client.update_leverage(
                market_index=market_id, margin_mode=margin_mode_int, leverage=leverage
            )

            # Result format: (created_tx, response, error)
            _, _, error = result

            if error:
                logger.error(f"Failed to update leverage for {instrument.symbol}: {error}")
                return False, error

            logger.info(f"Successfully updated leverage for {instrument.symbol} to {leverage}x ({margin_mode})")
            return True, None

        except Exception as e:
            error_msg = f"Exception updating leverage: {str(e)}"
            logger.error(error_msg)
            return False, error_msg

    async def transfer(
        self, to_account_index: int, amount: float, fee: float = 0.0, memo: str = ""
    ) -> tuple[bool, str | None]:
        """
        Transfer USDC between accounts.

        Args:
            to_account_index: Destination account index
            amount: Amount of USDC to transfer
            fee: Transfer fee (default: 0.0)
            memo: Optional memo/note for the transfer

        Returns:
            Tuple of (success: bool, error: str | None)

        Example:
            >>> # Transfer 100 USDC to account 225672
            >>> broker = ctx.get_broker("LIGHTER")
            >>> success, error = await broker.extensions.transfer(
            >>>     to_account_index=225672,
            >>>     amount=100.0,
            >>>     memo="Test transfer"
            >>> )
            >>> if success:
            >>>     print("Transfer successful")
            >>> else:
            >>>     print(f"Transfer failed: {error}")
        """
        try:
            logger.info(f"Transferring {amount} USDC to account {to_account_index} (fee={fee}, memo='{memo}')")

            # Use SignerClient to transfer
            # Note: eth_private_key is already stored in signer_client
            from lighter.signer_client import SignerClient

            result = await self.client.signer_client.transfer(
                eth_private_key=self.client.private_key,
                to_account_index=to_account_index,
                asset_id=SignerClient.ASSET_ID_USDC,  # 3 for USDC
                route_from=SignerClient.ROUTE_PERP,  # 0 for perp
                route_to=SignerClient.ROUTE_PERP,  # 0 for perp
                amount=amount,
                fee=int(fee),
                memo=memo,
            )

            # Result format: (created_tx, response, error)
            _, _, error = result

            if error:
                logger.error(f"Failed to transfer USDC: {error}")
                return False, error

            logger.info(f"Successfully transferred {amount} USDC to account {to_account_index}")
            return True, None

        except Exception as e:
            error_msg = f"Exception during transfer: {str(e)}"
            logger.error(error_msg)
            return False, error_msg

    async def create_pool(
        self, operator_fee: float, initial_total_shares: float, min_operator_share_rate: float
    ) -> tuple[bool, str | None]:
        """
        Create a public liquidity pool (Lighter-specific feature).

        Args:
            operator_fee: Fee charged by pool operator (as fraction, e.g., 0.01 for 1%)
            initial_total_shares: Initial number of shares to create
            min_operator_share_rate: Minimum operator share rate (as fraction)

        Returns:
            Tuple of (success: bool, error: str | None)

        Example:
            >>> # Create a pool with 1% operator fee
            >>> broker = ctx.get_broker("LIGHTER")
            >>> success, error = await broker.extensions.create_pool(
            >>>     operator_fee=0.01,
            >>>     initial_total_shares=1000000.0,
            >>>     min_operator_share_rate=0.05
            >>> )
            >>> if success:
            >>>     print("Pool created successfully")
            >>> else:
            >>>     print(f"Failed to create pool: {error}")
        """
        try:
            logger.info(
                f"Creating public pool: operator_fee={operator_fee}, "
                f"initial_shares={initial_total_shares}, min_operator_rate={min_operator_share_rate}"
            )

            # Use SignerClient to create pool
            result = await self.client.signer_client.create_public_pool(
                operator_fee=operator_fee,
                initial_total_shares=initial_total_shares,
                min_operator_share_rate=min_operator_share_rate,
            )

            # Result format: (created_tx, response, error)
            _, _, error = result

            if error:
                logger.error(f"Failed to create public pool: {error}")
                return False, error

            logger.info("Successfully created public pool")
            return True, None

        except Exception as e:
            error_msg = f"Exception creating pool: {str(e)}"
            logger.error(error_msg)
            return False, error_msg

    async def create_subaccount(self) -> tuple[bool, str | None]:
        """
        Create a new sub-account.

        Returns:
            Tuple of (success: bool, error: str | None)

        Example:
            >>> # Create a new sub-account
            >>> broker = ctx.get_broker("LIGHTER")
            >>> success, error = await broker.extensions.create_subaccount()
            >>> if success:
            >>>     print("Sub-account created successfully")
            >>> else:
            >>>     print(f"Failed to create sub-account: {error}")
        """
        try:
            logger.info("Creating new sub-account")

            # Use SignerClient to create sub-account
            result = await self.client.signer_client.create_sub_account()

            # Result format: (created_tx, response, error)
            _, _, error = result

            if error:
                logger.error(f"Failed to create sub-account: {error}")
                return False, error

            logger.info("Successfully created sub-account")
            return True, None

        except Exception as e:
            error_msg = f"Exception creating sub-account: {str(e)}"
            logger.error(error_msg)
            return False, error_msg



"""Market stats handler for Lighter WebSocket messages"""

import time
from typing import TYPE_CHECKING, Any

from qubx import logger
from qubx.core.basics import FundingPayment, FundingRate, Instrument, OpenInterest, dt_64
from qubx.utils.time import floor_t64

from .base import BaseHandler

if TYPE_CHECKING:
    from ..instruments import LighterInstrumentLoader


class MarketStatsHandler(BaseHandler[dict[Instrument, list[FundingRate | OpenInterest | FundingPayment]]]):
    """
    Handler for market stats messages.
    It creates open interest, funding rate, and funding payment data objects.

    Message format (example):
    ```json
    {
        'channel': 'market_stats:all',
        'market_stats': {'0': {'market_id': 0,
        'index_price': '4021.32',
        'mark_price': '4019.38',
        'open_interest': '176612968.404318',
        'open_interest_limit': '72057594037927936.000000',
        'funding_clamp_small': '0.0500',
        'funding_clamp_big': '4.0000',
        'last_trade_price': '4019.04',
        'current_funding_rate': '0.0004',
        'funding_rate': '-0.0002',
        'funding_timestamp': 1760972400002,
        'daily_base_token_volume': 377551.1336,
        'daily_quote_token_volume': 1515105897.814945,
        'daily_price_low': 3907.78,
        'daily_price_high': 4082.24,
        'daily_price_change': 0.9686203189212543}},
        'type': 'update/market_stats'
    }
    ```

    Returns:
        dict[Instrument, list[FundingRate | OpenInterest | FundingPayment]]
        Maps each instrument to list of data objects generated from the message.

    Note on funding rates:
        - 'current_funding_rate': Active rate (used for FundingRate.rate)
        - 'funding_rate': Rate that was paid at last funding_timestamp
        - 'funding_timestamp': When last payment occurred (in the past)
        - Lighter uses 1-hour funding intervals
    """

    FUNDING_INTERVAL_HOURS = 1  # Lighter uses 1-hour funding
    FUNDING_RATE_INTERVAL = "1min"
    OPEN_INTEREST_INTERVAL = "5min"

    def __init__(self, instrument_loader: "LighterInstrumentLoader"):
        """
        Initialize market stats handler.

        Args:
            instrument_loader: LighterInstrumentLoader for market_id → Instrument mapping
        """
        super().__init__()
        self.instrument_loader: LighterInstrumentLoader = instrument_loader

        # Track funding state per market_id for payment detection
        self._last_funding_timestamps: dict[int, int] = {}  # market_id → funding_timestamp (ms)

        # Buffer for interval-based emission (stores last message in current interval)
        self._funding_rate_buffer: dict[int, dict] = {}  # market_id → {data, raw_time}
        self._open_interest_buffer: dict[int, dict] = {}  # market_id → {data, raw_time}

        # Track last floored boundary we emitted (None = no emission yet)
        self._funding_rate_last_boundary: dict[int, dt_64 | None] = {}
        self._open_interest_last_boundary: dict[int, dt_64 | None] = {}

    def can_handle(self, message: dict[str, Any]) -> bool:
        """Check if message is market_stats update"""
        return (
            message.get("channel", "").startswith("market_stats:")
            and message.get("type") in ["update/market_stats", "subscribed/market_stats"]
            and "market_stats" in message
        )

    def _handle_impl(
        self, message: dict[str, Any]
    ) -> dict[Instrument, list[FundingRate | OpenInterest | FundingPayment]] | None:
        """
        Process all markets in message.

        Args:
            message: Raw Lighter market stats message

        Returns:
            Dictionary mapping Instrument to list of data objects, or None if no valid data

        Raises:
            ValueError: If message format is invalid
        """
        market_stats = message.get("market_stats")
        if not market_stats:
            raise ValueError("Missing market_stats in message")

        # Get message timestamp (use current time if not in message)
        msg_timestamp_ms = message.get("timestamp")
        if msg_timestamp_ms:
            msg_time_ns = int(msg_timestamp_ms * 1_000_000)  # ms → ns
            msg_time = dt_64(msg_time_ns, "ns")
        else:
            msg_time_ns = int(time.time() * 1_000_000_000)
            msg_time = dt_64(msg_time_ns, "ns")

        results: dict[Instrument, list[FundingRate | OpenInterest | FundingPayment]] = {}

        # Process each market in the message
        for market_id_str, market_data in market_stats.items():
            try:
                market_id = int(market_id_str)
            except ValueError:
                logger.warning(f"Invalid market_id format: {market_id_str}, skipping")
                continue

            # Resolve market_id → Instrument
            instrument = self.instrument_loader.get_instrument_by_market_id(market_id)
            if instrument is None:
                logger.debug(f"Unknown market_id: {market_id}, skipping")
                continue

            # Process this market's data
            objects = self._process_market_data(market_id, instrument, market_data, msg_time)

            if objects:
                results[instrument] = objects

        return results if results else None

    def _process_market_data(
        self, market_id: int, instrument: Instrument, data: dict, msg_time: dt_64
    ) -> list[FundingRate | OpenInterest | FundingPayment]:
        """
        Process single market's stats data.

        Args:
            market_id: Lighter market ID
            instrument: Qubx Instrument
            data: Market stats data from message
            msg_time: Message timestamp as dt_64

        Returns:
            List of data objects (FundingRate, OpenInterest, FundingPayment)
        """
        objects: list[FundingRate | OpenInterest | FundingPayment] = []

        try:
            # Extract fields (all are strings in message)
            funding_rate_str = data.get("funding_rate")  # Past payment rate
            current_funding_rate_str = data.get("current_funding_rate")  # Active rate
            funding_timestamp_ms = data.get("funding_timestamp")  # When last payment occurred
            open_interest_str = data.get("open_interest")
            mark_price_str = data.get("mark_price")
            index_price_str = data.get("index_price")

            # Convert to appropriate types
            current_funding_rate = float(current_funding_rate_str) if current_funding_rate_str else None
            funding_timestamp = int(funding_timestamp_ms) if funding_timestamp_ms else None
            mark_price = float(mark_price_str) if mark_price_str else None
            index_price = float(index_price_str) if index_price_str else None

            # 1. Create FundingRate (active rate) - interval-based emission
            if current_funding_rate is not None and funding_timestamp is not None:
                # funding_timestamp is when LAST payment occurred
                # next payment is funding_timestamp + 1 hour
                funding_timestamp_ns = int(funding_timestamp * 1_000_000)
                next_funding_ns = funding_timestamp_ns + (self.FUNDING_INTERVAL_HOURS * 3600 * 1_000_000_000)

                # Prepare data for buffering
                funding_rate_data = {
                    "rate": current_funding_rate / 100.0,  # Convert from percentage to decimal
                    "next_funding_time": dt_64(next_funding_ns, "ns"),
                    "mark_price": mark_price,
                    "index_price": index_price,
                }

                # Check if we should emit based on interval
                should_emit, emission_time, data_to_emit = self._check_interval_emission(
                    market_id,
                    msg_time,
                    self._funding_rate_buffer,
                    self._funding_rate_last_boundary,
                    self.FUNDING_RATE_INTERVAL,
                )

                if should_emit and data_to_emit:
                    assert emission_time is not None, "emission_time should not be None when should_emit is True"
                    # Emit with boundary timestamp
                    funding_rate_obj = FundingRate(
                        time=emission_time,  # Boundary timestamp (when data becomes available)
                        rate=data_to_emit["rate"],
                        interval="1h",
                        next_funding_time=data_to_emit["next_funding_time"],
                        mark_price=data_to_emit.get("mark_price"),
                        index_price=data_to_emit.get("index_price"),
                    )
                    objects.append(funding_rate_obj)

                # Always buffer current data
                self._funding_rate_buffer[market_id] = {
                    "data": funding_rate_data,
                    "raw_time": msg_time,
                }

            # 2. Create OpenInterest - interval-based emission
            if open_interest_str and mark_price:
                open_interest = float(open_interest_str)
                open_interest_usd = open_interest * mark_price

                # Prepare data for buffering
                oi_data = {
                    "symbol": instrument.symbol,
                    "open_interest": open_interest,
                    "open_interest_usd": open_interest_usd,
                }

                # Check if we should emit based on interval
                should_emit, emission_time, data_to_emit = self._check_interval_emission(
                    market_id,
                    msg_time,
                    self._open_interest_buffer,
                    self._open_interest_last_boundary,
                    self.OPEN_INTEREST_INTERVAL,
                )

                if should_emit and data_to_emit:
                    assert emission_time is not None, "emission_time should not be None when should_emit is True"
                    # Emit with boundary timestamp
                    oi_obj = OpenInterest(
                        time=emission_time,  # Boundary timestamp (when data becomes available)
                        symbol=data_to_emit["symbol"],
                        open_interest=data_to_emit["open_interest"],
                        open_interest_usd=data_to_emit["open_interest_usd"],
                    )
                    objects.append(oi_obj)

                # Always buffer current data
                self._open_interest_buffer[market_id] = {
                    "data": oi_data,
                    "raw_time": msg_time,
                }

            # 3. Detect and create FundingPayment (when funding_timestamp changes)
            if funding_timestamp is not None and funding_rate_str is not None:
                if self._should_emit_funding_payment(market_id, funding_timestamp):
                    funding_rate_paid = float(funding_rate_str)
                    funding_timestamp_ns = int(funding_timestamp * 1_000_000)

                    payment = FundingPayment(
                        time=floor_t64(
                            dt_64(funding_timestamp_ns, "ns"), self.FUNDING_INTERVAL_HOURS
                        ),  # When payment occurred
                        funding_rate=funding_rate_paid / 100.0,  # Convert from percentage to decimal
                        funding_interval_hours=self.FUNDING_INTERVAL_HOURS,
                    )
                    objects.append(payment)

                # Update tracking
                self._last_funding_timestamps[market_id] = funding_timestamp

        except (ValueError, TypeError) as e:
            logger.error(f"Error processing market {market_id} ({instrument.symbol}) stats: {e}")
            return []

        return objects

    def _should_emit_funding_payment(self, market_id: int, current_funding_timestamp: int) -> bool:
        """
        Check if funding payment should be emitted.
        Emit when funding_timestamp advances (new hour started).

        Args:
            market_id: Market ID
            current_funding_timestamp: Current funding_timestamp from message (ms)

        Returns:
            True if payment should be emitted
        """
        last_timestamp = self._last_funding_timestamps.get(market_id)

        # Don't emit on first message (no previous state)
        if last_timestamp is None:
            return False

        # Emit if timestamp advanced (new funding period)
        return current_funding_timestamp > last_timestamp

    def _check_interval_emission(
        self,
        market_id: int,
        msg_time: dt_64,
        buffer_dict: dict[int, dict],
        last_boundary_dict: dict[int, dt_64 | None],
        interval: str,
    ) -> tuple[bool, dt_64 | None, dict | None]:
        """
        Check if we should emit based on interval boundary crossing.

        This implements interval-based emission with no lookahead bias:
        - First message: buffer only, don't emit
        - Boundary crossed: emit buffered data from previous interval with current boundary timestamp
        - Same interval: buffer only (overwrites previous)

        Args:
            market_id: Market ID
            msg_time: Current message timestamp
            buffer_dict: Buffer dictionary to check/update
            last_boundary_dict: Last boundary dictionary to check/update
            interval: Interval string (e.g., "1min", "5min")

        Returns:
            (should_emit, emission_timestamp, data_to_emit)
        """
        # Floor current message time to interval
        current_boundary = floor_t64(msg_time, interval)

        # Get last emitted boundary
        last_boundary = last_boundary_dict.get(market_id)

        # First message ever - buffer but don't emit
        if last_boundary is None:
            last_boundary_dict[market_id] = current_boundary
            return (False, None, None)

        # Check if we crossed a boundary
        if current_boundary > last_boundary:
            # Get buffered data from PREVIOUS interval
            buffered = buffer_dict.get(market_id)

            if buffered is None:
                # No data to emit (shouldn't happen, but handle gracefully)
                last_boundary_dict[market_id] = current_boundary
                return (False, None, None)

            # Emit buffered data with CURRENT boundary timestamp
            # (This is when the data becomes available, not when it was measured)
            last_boundary_dict[market_id] = current_boundary
            return (True, current_boundary, buffered["data"])

        # Still in same interval - buffer only
        return (False, None, None)



import configparser
import dataclasses
import glob
import json
import os
import re
from datetime import datetime
from pathlib import Path

import pandas as pd
import stackprinter

from qubx import logger
from qubx.core.basics import (
    ZERO_COSTS,
    AssetType,
    FeesLookup,
    Instrument,
    InstrumentsLookup,
    MarketType,
    TransactionCostsCalculator,
)
from qubx.utils.marketdata.dukas import SAMPLE_INSTRUMENTS
from qubx.utils.misc import get_local_qubx_folder, load_qubx_resources_as_json, load_qubx_resources_as_text, makedirs

_DEF_INSTRUMENTS_FOLDER = "instruments"
_DEF_FEES_FOLDER = "fees"

_PACKAGED_FEES_FILE = "crypto-fees.ini"

_INI_FILE = "settings.ini"
_INI_SECTION_INSTRUMENTS = "instrument-lookup"
_INI_SECTION_FEES = "fees-lookup"


class _InstrumentEncoder(json.JSONEncoder):
    def default(self, obj):
        if dataclasses.is_dataclass(obj):
            return {k: v for k, v in dataclasses.asdict(obj).items() if not k.startswith("_")}
        if isinstance(obj, (datetime)):
            return obj.isoformat()
        return super().default(obj)


class _InstrumentDecoder(json.JSONDecoder):
    def decode(self, json_string):
        obj = super(_InstrumentDecoder, self).decode(json_string)
        if isinstance(obj, dict):
            # Convert delivery_date and onboard_date strings to datetime
            if (delivery_date := obj.get("delivery_date")) and delivery_date != "NaT":
                obj["delivery_date"] = pd.Timestamp(delivery_date)

            if (onboard_date := obj.get("onboard_date")) and onboard_date != "NaT":
                obj["onboard_date"] = pd.Timestamp(onboard_date)

            if (delist_date := obj.get("delist_date")) and delist_date != "NaT":
                obj["delist_date"] = pd.Timestamp(delist_date)

            return Instrument(
                symbol=obj["symbol"],
                asset_type=AssetType[obj["asset_type"]],
                market_type=MarketType[obj["market_type"]],
                exchange=obj["exchange"],
                base=obj["base"],
                quote=obj["quote"],
                settle=obj["settle"],
                exchange_symbol=obj.get("exchange_symbol", obj["symbol"]),
                tick_size=float(obj["tick_size"]),
                lot_size=float(obj["lot_size"]),
                min_size=float(obj["min_size"]),
                min_notional=float(obj.get("min_notional", 0.0)),
                initial_margin=float(obj.get("initial_margin", 0.0)),
                maint_margin=float(obj.get("maint_margin", 0.0)),
                liquidation_fee=float(obj.get("liquidation_fee", 0.0)),
                contract_size=float(obj.get("contract_size", 1.0)),
                onboard_date=obj.get("onboard_date", None),
                delivery_date=obj.get("delivery_date", None),
                inverse=obj.get("inverse", False),
                delist_date=obj.get("delist_date", None),
            )
        elif isinstance(obj, list):
            return [self.decode(json.dumps(item)) for item in obj]
        return obj


class FileInstrumentsLookupWithCCXT(InstrumentsLookup):
    _lookup: dict[str, Instrument]
    _path: str

    def __init__(
        self, path: str = makedirs(get_local_qubx_folder(), _DEF_INSTRUMENTS_FOLDER), query_exchanges=False
    ) -> None:
        self._path = path
        if not self.load():
            self.refresh(query_exchanges)
        self.load()

    def load(self) -> bool:
        self._lookup = {}
        data_exists = False
        for fs in glob.glob(self._path + "/*.json"):
            try:
                with open(fs, "r") as f:
                    for i in json.load(f, cls=_InstrumentDecoder):
                        self._lookup[f"{i.exchange}:{i.market_type}:{i.symbol}"] = i
                    data_exists = True
            except Exception as ex:
                stackprinter.show_current_exception()
                logger.warning(ex)

        return data_exists

    def get_lookup(self) -> dict[str, Instrument]:
        return self._lookup

    def _save_to_json(self, path, instruments: list[Instrument]):
        with open(path, "w") as f:
            json.dump(instruments, f, cls=_InstrumentEncoder, indent=4)
        logger.info(f"Saved {len(instruments)} to {path}")

    def refresh(self, query_exchanges: bool = False):
        for mn in dir(self):
            if mn.startswith("_update_"):
                getattr(self, mn)(self._path, query_exchanges)

    def _copy_instruments_and_update_from_ccxt(
        self,
        path: str,
        file_name: str,
        exchange_to_ccxt_name: dict[str, str],
        keep_types: list[MarketType] | None = None,
        query_exchanges: bool = False,
    ):
        from qubx.utils.marketdata.ccxt import ccxt_fetch_instruments

        # - first we try to load packed data from QUBX resources
        instruments = {}
        try:
            _package_data = load_qubx_resources_as_json(f"instruments/symbols-{file_name}")
            if _package_data:
                for i in _convert_instruments_metadata_to_qubx(_package_data):
                    instruments[i] = i
        except Exception as e:
            logger.warning(f"Can't load resource file from instruments/symbols-{file_name} - {str(e)}")

        if query_exchanges:
            # - replace defaults with data from CCXT
            instruments = ccxt_fetch_instruments(exchange_to_ccxt_name, keep_types, instruments)

        # - save to file
        self._save_to_json(os.path.join(path, f"{file_name}.json"), list(instruments.values()))

    def _update_kraken(self, path: str, query_exchanges: bool = False):
        self._copy_instruments_and_update_from_ccxt(
            path, "kraken-spot", {"kraken": "kraken"}, keep_types=[MarketType.SPOT], query_exchanges=query_exchanges
        )
        self._copy_instruments_and_update_from_ccxt(
            path,
            "kraken.f-perpetual",
            {"kraken.f": "krakenfutures"},
            keep_types=[MarketType.SWAP],
            query_exchanges=query_exchanges,
        )
        self._copy_instruments_and_update_from_ccxt(
            path,
            "kraken.f-future",
            {"kraken.f": "krakenfutures"},
            keep_types=[MarketType.FUTURE],
            query_exchanges=query_exchanges,
        )

    def _update_hyperliquid(self, path: str, query_exchanges: bool = False):
        self._copy_instruments_and_update_from_ccxt(
            path,
            "hyperliquid-spot",
            {"hyperliquid": "hyperliquid"},
            keep_types=[MarketType.SPOT],
            query_exchanges=query_exchanges,
        )
        self._copy_instruments_and_update_from_ccxt(
            path,
            "hyperliquid.f-perpetual",
            {"hyperliquid.f": "hyperliquid"},
            keep_types=[MarketType.SWAP],
            query_exchanges=query_exchanges,
        )

    def _update_binance(self, path: str, query_exchanges: bool = False):
        self._copy_instruments_and_update_from_ccxt(
            path,
            "binance-spot",
            {"binance": "binance"},
            keep_types=[MarketType.SPOT, MarketType.MARGIN],
            query_exchanges=query_exchanges,
        )
        self._copy_instruments_and_update_from_ccxt(
            path,
            "binance.um-perpetual",
            {"binance.um": "binanceusdm"},
            keep_types=[MarketType.SWAP],
            query_exchanges=query_exchanges,
        )

        self._copy_instruments_and_update_from_ccxt(
            path,
            "binance.um-future",
            {"binance.um": "binanceusdm"},
            keep_types=[MarketType.FUTURE],
            query_exchanges=query_exchanges,
        )
        self._copy_instruments_and_update_from_ccxt(
            path,
            "binance.cm-perpetual",
            {"binance.cm": "binancecoinm"},
            keep_types=[MarketType.SWAP],
            query_exchanges=query_exchanges,
        )
        self._copy_instruments_and_update_from_ccxt(
            path,
            "binance.cm-future",
            {"binance.cm": "binancecoinm"},
            keep_types=[MarketType.FUTURE],
            query_exchanges=query_exchanges,
        )

    # todo: temporaty disabled ccxt call to exchange, due to conectivity issues. Revert for bitfinex live usage
    def _update_bitfinex(self, path: str, query_exchanges: bool = False):
        self._copy_instruments_and_update_from_ccxt(
            path,
            "bitfinex.f-perpetual",
            {"bitfinex.f": "bitfinex"},
            keep_types=[MarketType.SWAP],
            query_exchanges=False,
        )

    def _update_bitmex(self, path: str, query_exchanges: bool = False):
        self._copy_instruments_and_update_from_ccxt(
            path,
            "bitmex",
            {"bitmex": "bitmex"},
            query_exchanges=query_exchanges,
        )

    def _update_deribit(self, path: str, query_exchanges: bool = False):
        self._copy_instruments_and_update_from_ccxt(
            path,
            "deribit",
            {"deribit": "deribit"},
            keep_types=[MarketType.SWAP],
            query_exchanges=query_exchanges,
        )

    def _update_bybit(self, path: str, query_exchanges: bool = False):
        self._copy_instruments_and_update_from_ccxt(
            path,
            "bybit.f",
            {"bybit.f": "bybit"},
            keep_types=[MarketType.SWAP],
            query_exchanges=query_exchanges,
        )

    def _update_okx(self, path: str, query_exchanges: bool = False):
        self._copy_instruments_and_update_from_ccxt(
            path,
            "okx.f",
            {"okx.f": "okx"},
            keep_types=[MarketType.SWAP],
            query_exchanges=query_exchanges,
        )

    def _update_dukas(self, path: str, query_exchanges: bool = False):
        self._save_to_json(os.path.join(path, "dukas.json"), SAMPLE_INSTRUMENTS)


class FeesLookupFile(FeesLookup):
    """
    Fees lookup
    """

    _lookup: dict[str, tuple[float, float]]
    _path: str

    def __init__(self, path: str = makedirs(get_local_qubx_folder(), _DEF_FEES_FOLDER)) -> None:
        self._path = path
        if not self.load():
            self.refresh()
        self.load()

    def load(self) -> bool:
        self._lookup = {}
        data_exists = False
        parser = configparser.ConfigParser()
        # - load all avaliable configs
        for fs in glob.glob(self._path + "/*.ini"):
            parser.read(fs)
            data_exists = True

        for exch in parser.sections():
            for spec, info in parser[exch].items():
                try:
                    maker, taker = info.split(",")
                    self._lookup[f"{exch}_{spec}"] = (float(maker), float(taker))
                except (ValueError, TypeError) as e:
                    logger.warning(f'Wrong spec format for {exch}: "{info}". Should be spec=maker,taker. Error: {e}')

        return data_exists

    def refresh(self):
        try:
            _packaged_fees = load_qubx_resources_as_text(_PACKAGED_FEES_FILE)
            with open(os.path.join(self._path, "default.ini"), "w") as f:
                f.write(_packaged_fees)
        except Exception as e:
            logger.error(f"Can't load resource file from {_PACKAGED_FEES_FILE} - {str(e)}")

    def find_fees(self, exchange: str, spec: str | None) -> TransactionCostsCalculator:
        if spec is None:
            return ZERO_COSTS

        key = f"{exchange.lower()}_{spec}"

        # - check if spec is of type maker=...,taker=...
        # Check if spec is in the format maker=X,taker=Y
        maker_taker_pattern = re.compile(r"maker=(-?[0-9.]+)[,\s]taker=(-?[0-9.]+)")
        match = maker_taker_pattern.match(spec)
        if match:
            maker_rate, taker_rate = float(match.group(1)), float(match.group(2))
            return TransactionCostsCalculator(key, maker_rate, taker_rate)

        # - otherwise lookup in lookup table
        vals = self._lookup.get(key)
        if vals is None:
            raise ValueError(f"No fees found for {key}")

        assert isinstance(vals, tuple)
        return TransactionCostsCalculator(key, vals[0], vals[1])

    def __repr__(self) -> str:
        s = "Name:\t\t\t(maker, taker)\n"
        for k, v in self._lookup.items():
            s += f"{k.ljust(25)}: {v}\n"
        return s


def _convert_instruments_metadata_to_qubx(data: list[dict]) -> list[Instrument]:
    """
    Converting tardis symbols meta-data to Qubx instruments
    """
    _excs = {
        "binance": "BINANCE",
        "binance-delivery": "BINANCE.CM",
        "binance-futures": "BINANCE.UM",
        "kraken": "KRAKEN",
        "cryptofacilities": "KRAKEN.F",
        "bitfinex": "BITFINEX",
        "bitfinex-derivatives": "BITFINEX.F",
        "hyperliquid": "HYPERLIQUID",
    }
    r = []
    for s in data:
        _pfx = ""
        if _delist_date := s.get("availableTo", None):
            _delist_date = pd.Timestamp(_delist_date)

        if _delivery_date := s.get("expiry", None):
            _delivery_date = pd.Timestamp(_delivery_date)

        match s["type"]:
            case "perpetual":
                _type = MarketType.SWAP
            case "spot":
                _type = MarketType.SPOT
            case "future":
                _type = MarketType.FUTURE
                if _delivery_date:
                    _pfx = "." + _delivery_date.strftime("%Y%m%d")
            case _:
                raise ValueError(f" -> Unsupported type {s['type']}")
        r.append(
            Instrument(
                s["baseCurrency"] + s["quoteCurrency"] + _pfx,
                AssetType.CRYPTO,
                _type,
                _excs.get(s["exchange"], s["exchange"].upper()),
                s["baseCurrency"],
                s["quoteCurrency"],
                s["quoteCurrency"],
                s["datasetId"],
                tick_size=s["priceIncrement"],
                lot_size=s["minTradeAmount"],
                min_size=s["amountIncrement"],
                min_notional=0,  # we don't have this info from tardis
                contract_size=s.get("contractMultiplier", 1.0),
                onboard_date=s.get("availableSince", None),
                delivery_date=_delivery_date,
                inverse=s.get("inverse", False),
                delist_date=_delist_date,
            )
        )
    return r


class InstrumentsLookupMongo(InstrumentsLookup):
    _MONGO_DB_BASE_NAME = "metadata"
    _MONGO_DB_TABLE_NAME = "instruments"

    _lookup: dict[str, Instrument]
    _mongo_url: str
    _reload_interval: pd.Timedelta | None
    _last_refresh: pd.Timestamp

    def __init__(self, mongo_url: str = "mongodb://localhost:27017/", reload_interval: str | None = None):
        self._mongo_url = mongo_url
        self._reload_interval = pd.Timedelta(reload_interval) if reload_interval else None
        self.load()

    def load(self):
        from pymongo import MongoClient

        self._lookup = {}
        with MongoClient(self._mongo_url) as client:
            db = client[self._MONGO_DB_BASE_NAME]
            collection = db[self._MONGO_DB_TABLE_NAME]
            for i in collection.find():
                i.pop("_id")
                instr = Instrument(**i)
                self._lookup[f"{instr.exchange}:{instr.market_type}:{instr.symbol}"] = instr

        self._last_refresh = pd.Timestamp.now()

    def get_lookup(self) -> dict[str, Instrument]:
        # - reload data if needed
        if self._reload_interval and pd.Timestamp.now() - self._last_refresh > self._reload_interval:
            self.load()

        return self._lookup


class LookupsManager(InstrumentsLookup, FeesLookup):
    _i_lookup: InstrumentsLookup
    _t_lookup: FeesLookup

    def __new__(cls):
        if not hasattr(cls, "instance"):
            cls.instance = super(LookupsManager, cls).__new__(cls)

            # - try to load settings
            parser = configparser.ConfigParser()
            parser.read(Path(get_local_qubx_folder()) / _INI_FILE)

            if _INI_SECTION_INSTRUMENTS in parser:
                cls.instance._i_lookup = LookupsManager._get_instrument_lookup(**dict(parser[_INI_SECTION_INSTRUMENTS]))
            else:
                cls.instance._i_lookup = FileInstrumentsLookupWithCCXT()

            if _INI_SECTION_FEES in parser:
                cls.instance._t_lookup = LookupsManager._get_fees_lookup(**dict(parser[_INI_SECTION_FEES]))
            else:
                cls.instance._t_lookup = FeesLookupFile()

        return cls.instance

    @staticmethod
    def _get_instrument_lookup(type: str, **kwargs) -> InstrumentsLookup:
        match type.lower():
            case "file":
                return FileInstrumentsLookupWithCCXT(**kwargs)
            case "mongo":
                return InstrumentsLookupMongo(**kwargs)
            case _:
                raise ValueError(f"Invalid lookup type: {type}")

    @staticmethod
    def _get_fees_lookup(type: str, **kwargs) -> FeesLookup:
        match type.lower():
            case "file":
                return FeesLookupFile(**kwargs)
            case _:
                raise ValueError(f"Invalid lookup type: {type}")

    def find_symbol(self, exchange: str, symbol: str, market_type: MarketType | None = None) -> Instrument | None:
        return self._i_lookup.find_symbol(exchange, symbol, market_type)

    def find_instruments(
        self,
        exchange: str,
        base: str | None = None,
        quote: str | None = None,
        market_type: MarketType | None = None,
        as_of: str | pd.Timestamp | None = None,
    ) -> list[Instrument]:
        return self._i_lookup.find_instruments(exchange, base, quote, market_type, as_of=as_of)

    def find_aux_instrument_for(
        self, instrument: Instrument, base_currency: str, market_type: MarketType | None = None
    ) -> Instrument | None:
        return self._i_lookup.find_aux_instrument_for(instrument, base_currency, market_type)

    def find_fees(self, exchange: str, spec: str | None) -> TransactionCostsCalculator:
        return self._t_lookup.find_fees(exchange, spec)

    def __getitem__(self, spath: str) -> list[Instrument]:
        return self._i_lookup[spath]


# - global lookup helper
lookup = LookupsManager()



from typing import Any

import pandas as pd
from rich.console import RenderableType
from rich.table import Table
from textual import on
from textual.app import App, ComposeResult
from textual.containers import Horizontal, Vertical
from textual.widgets import (
    Button,
    DataTable,
    Footer,
    Header,
    Label,
    Static,
    Tree,
)

from qubx import logger
from qubx.backtester import BacktestsResultsManager
from qubx.core.metrics import TradingSessionResult, get_cum_pnl


class BacktestTreeNode:
    """Represents a node in the backtest tree structure"""

    def __init__(self, name: str, path: str | None = None, manager: BacktestsResultsManager | None = None):
        self.name = name
        self.path = path
        self.manager = manager
        self.children: list[BacktestTreeNode] = []
        self.backtest_results: list[dict[str, Any]] = []
        self.is_leaf = False

    def add_child(self, child: "BacktestTreeNode"):
        self.children.append(child)

    def add_backtest(self, result_info: dict[str, Any]):
        self.backtest_results.append(result_info)
        self.is_leaf = True


class BacktestResultsTree(Tree[BacktestTreeNode]):
    """Custom tree widget for backtest results"""

    def __init__(self, root_path: str):
        self.root_path = root_path
        self.manager = BacktestsResultsManager(root_path)

        # Build tree structure from backtest results
        root_node = BacktestTreeNode("Backtest Results", root_path, self.manager)
        self._build_tree_structure(root_node)

        super().__init__(root_node.name, data=root_node)
        self._populate_tree(self.root, root_node)

    def _build_tree_structure(self, root_node: BacktestTreeNode):
        """Build tree structure from backtest manager results"""
        # Group by strategy class for organization
        strategy_groups: dict[str, list[dict]] = {}

        for name, info in self.manager.results.items():
            strategy_class = info.get("strategy_class", "Unknown").split(".")[-1]
            if strategy_class not in strategy_groups:
                strategy_groups[strategy_class] = []
            info["display_name"] = name
            strategy_groups[strategy_class].append(info)

        # Create strategy nodes
        for strategy_name, results in strategy_groups.items():
            strategy_node = BacktestTreeNode(f"{strategy_name} ({len(results)} results)", strategy_name, self.manager)
            root_node.add_child(strategy_node)

            for result in results:
                strategy_node.add_backtest(result)

        # Add variations if any
        for idx, var_info in self.manager.variations.items():
            var_node = BacktestTreeNode(
                f"📊 {var_info['name']} ({len(var_info['variations'])} variations)", var_info["name"], self.manager
            )
            root_node.add_child(var_node)

            for variation in var_info["variations"]:
                var_node.add_backtest(variation)

    def _populate_tree(self, tree_node, data_node: BacktestTreeNode):
        """Populate the textual tree with data nodes"""
        for child in data_node.children:
            child_tree_node = tree_node.add(child.name, data=child)
            if child.children:
                self._populate_tree(child_tree_node, child)


class MetricsTable(DataTable):
    """Data table for displaying backtest metrics with row selection"""

    def __init__(self):
        super().__init__()
        self.zebra_stripes = True
        self.cursor_type = "row"

    def populate_table(self, results_data: list[dict[str, Any]]):
        """Populate table with backtest results"""
        if not results_data:
            return

        # Clear existing data and columns
        self.clear(columns=True)

        # Define columns based on metrics
        columns = [
            "Name",
            "Strategy",
            "Created",
            "Author",
            "Gain",
            "CAGR",
            "Sharpe",
            "QR",
            "Max DD %",
            "MDD USD",
            "Fees",
            "Executions",
        ]

        # Add columns only if not already present
        for col in columns:
            column_key = col.lower().replace(" ", "_")
            self.add_column(col, key=column_key)

        # Add rows
        for result in results_data:
            performance = result.get("performance", {})
            row_data = [
                result.get("display_name", result.get("name", "Unknown")),
                result.get("strategy_class", "").split(".")[-1],
                pd.Timestamp(result.get("creation_time", "")).strftime("%Y-%m-%d %H:%M"),
                result.get("author", ""),
                f"{performance.get('gain', 0):.3f}",
                f"{performance.get('cagr', 0):.3f}",
                f"{performance.get('sharpe', 0):.3f}",
                f"{performance.get('qr', 0):.3f}",
                f"{performance.get('max_dd_pct', 0):.3f}",
                f"{performance.get('mdd_usd', 0):.3f}",
                f"{performance.get('fees', 0):.3f}",
                f"{performance.get('execs', 0):.0f}",
            ]
            self.add_row(*row_data, key=result.get("name", ""))

    def sort_by_column(self, column: str):
        """Sort table by specified column"""
        self.sort(column)


class EquityChart(Static):
    """Widget for displaying equity curves"""

    def __init__(self):
        super().__init__()
        self.results: list[TradingSessionResult] = []

    def set_results(self, results: list[TradingSessionResult]):
        """Set the trading results for chart display"""
        self.results = results
        self.refresh_chart()

    def refresh_chart(self):
        """Generate and display equity chart as text representation"""
        if not self.results:
            self.update("No results to display")
            return

        try:
            # Validate that all results are TradingSessionResult objects
            for i, result in enumerate(self.results):
                if not hasattr(result, "portfolio_log"):
                    logger.error(f"Result {i} ({type(result)}) does not have portfolio_log attribute")
                    self.update(f"Invalid result type at index {i}: {type(result)}")
                    return

            # Get equity data for all results
            equity_data = get_cum_pnl(self.results, timeframe="1h")

            if isinstance(equity_data, pd.Series):
                equity_data = equity_data.to_frame()

            # Create a simple text-based chart representation
            chart_text = self._create_text_chart(equity_data)
            self.update(chart_text)

        except Exception as e:
            logger.error(f"Failed to generate equity chart: {e}")
            self.update(f"Chart generation failed: {e}\nResults types: {[type(r) for r in self.results]}")

    def _create_text_chart(self, data: pd.DataFrame) -> RenderableType:
        """Create a simple text representation of equity curves"""
        if data.empty:
            return "No data available for chart"

        # Create a simple summary table instead of complex chart
        table = Table(title="Equity Summary")
        table.add_column("Strategy", style="cyan")
        table.add_column("Start Return", style="green")
        table.add_column("End Return", style="green")
        table.add_column("Max Return", style="green")
        table.add_column("Min Return", style="red")

        for col in data.columns:
            series = data[col].dropna()
            if not series.empty:
                table.add_row(
                    str(col),
                    f"{series.iloc[0]:.2%}" if len(series) > 0 else "N/A",
                    f"{series.iloc[-1]:.2%}" if len(series) > 0 else "N/A",
                    f"{series.max():.2%}" if len(series) > 0 else "N/A",
                    f"{series.min():.2%}" if len(series) > 0 else "N/A",
                )

        return table


class TearsheetViewer(Static):
    """Widget for displaying tearsheet information"""

    def __init__(self):
        super().__init__()
        self.update(
            "Select a backtest result from the table to view its tearsheet.\nClick on any row to generate and open the tearsheet in your browser."
        )

    def show_tearsheet_info(self, result_name: str, file_path: str):
        """Show information about generated tearsheet"""
        self.update(
            f"Tearsheet generated for: {result_name}\n\nFile saved to: {file_path}\n\nThe tearsheet has been opened in your default browser.\nIf it didn't open automatically, you can manually open the file above."
        )


class BacktestBrowserApp(App):
    """Main TUI application for browsing backtest results"""

    CSS = """
    .box {
        height: 1fr;
    }
    
    #tree-container {
        width: 30%;
        border: solid $primary;
    }
    
    #content-container {
        width: 70%;
        border: solid $primary;
    }
    
    #metrics-table {
        height: 70%;
    }
    
    #tearsheet-info {
        height: 30%;
    }
    
    #controls {
        height: 3;
        background: $surface;
    }
    """

    def __init__(self, root_path: str):
        super().__init__()
        self.root_path = root_path
        self.current_results: list[dict[str, Any]] = []
        self.current_manager: BacktestsResultsManager | None = None

    def compose(self) -> ComposeResult:
        """Compose the TUI layout"""
        yield Header()

        with Horizontal():
            with Vertical(id="tree-container", classes="box"):
                yield Label("Backtest Results Tree")
                yield BacktestResultsTree(self.root_path)

            with Vertical(id="content-container", classes="box"):
                with Horizontal(id="controls"):
                    yield Button("Sort by Sharpe", id="sort-sharpe")
                    yield Button("Sort by CAGR", id="sort-cagr")
                    yield Button("Sort by Gain", id="sort-gain")
                    yield Button("Refresh", id="refresh")

                table = MetricsTable()
                table.id = "metrics-table"
                yield table

                tearsheet_viewer = TearsheetViewer()
                tearsheet_viewer.id = "tearsheet-info"
                yield tearsheet_viewer

        yield Footer()

    def on_mount(self) -> None:
        """Initialize the app when mounted"""
        self.title = f"Qubx Backtest Browser - {self.root_path}"

    @on(Tree.NodeSelected)
    def handle_tree_selection(self, event: Tree.NodeSelected) -> None:
        """Handle tree node selection"""
        if event.node.data:
            node_data = event.node.data
            self.current_results = node_data.backtest_results
            self.current_manager = node_data.manager

            if self.current_results:
                self._update_table()

    @on(DataTable.RowSelected)
    def handle_row_selection(self, event: DataTable.RowSelected) -> None:
        """Handle table row selection to show tearsheet"""
        if not self.current_manager or not self.current_results:
            return

        # Get the selected row key (which corresponds to the result name)
        row_key = event.row_key.value if event.row_key else None
        if not row_key:
            return

        # Find the corresponding result info
        result_info = None
        for result in self.current_results:
            if result.get("name") == row_key:
                result_info = result
                break

        if not result_info:
            logger.warning(f"Could not find result info for row key: {row_key}")
            return

        self._generate_tearsheet(result_info)

    @on(Button.Pressed, "#sort-sharpe")
    def sort_by_sharpe(self) -> None:
        """Sort results by Sharpe ratio"""
        self._sort_table("sharpe")

    @on(Button.Pressed, "#sort-cagr")
    def sort_by_cagr(self) -> None:
        """Sort results by CAGR"""
        self._sort_table("cagr")

    @on(Button.Pressed, "#sort-gain")
    def sort_by_gain(self) -> None:
        """Sort results by total gain"""
        self._sort_table("gain")

    @on(Button.Pressed, "#refresh")
    def refresh_data(self) -> None:
        """Refresh the data from the backtest manager"""
        try:
            # Get the current tree widget and its container
            tree_container = self.query_one("#tree-container")
            old_tree = self.query_one(BacktestResultsTree)

            # Remove the old tree
            old_tree.remove()

            # Create a new tree with refreshed data
            new_tree = BacktestResultsTree(self.root_path)
            tree_container.mount(new_tree)

            # Clear current results to force reselection
            self.current_results = []
            self.current_manager = new_tree.manager

            # Clear the table
            table = self.query_one("#metrics-table", MetricsTable)
            table.clear(columns=True)

            # Update tearsheet viewer
            tearsheet_viewer = self.query_one("#tearsheet-info", TearsheetViewer)
            tearsheet_viewer.update("Data refreshed. Select a strategy from the tree to view results.")

        except Exception as e:
            logger.error(f"Failed to refresh data: {e}")
            tearsheet_viewer = self.query_one("#tearsheet-info", TearsheetViewer)
            tearsheet_viewer.update(f"Failed to refresh data: {e}")

    def _sort_table(self, metric: str) -> None:
        """Sort table by specified metric"""
        if self.current_results:
            # Use DataTable's built-in sorting by column key
            column_key = metric.lower().replace(" ", "_")
            table = self.query_one("#metrics-table", MetricsTable)
            try:
                table.sort(column_key, reverse=True)  # Sort descending for better metrics first
            except Exception as e:
                logger.warning(f"Failed to sort by {metric}: {e}")
                # Fallback to manual sorting and repopulation
                self.current_results.sort(key=lambda x: x.get("performance", {}).get(metric, 0), reverse=True)
                self._update_table()

    def _update_table(self) -> None:
        """Update the metrics table"""
        table = self.query_one("#metrics-table", MetricsTable)
        table.populate_table(self.current_results)

    def _generate_tearsheet(self, result_info: dict[str, Any]) -> None:
        """Generate tearsheet for selected result"""
        try:
            import tempfile
            import webbrowser

            name = result_info.get("name", "")
            display_name = result_info.get("display_name", name)

            if not name or not self.current_manager:
                return

            # Load the trading result
            trading_result = self.current_manager.load(name)
            if isinstance(trading_result, list):
                trading_result = trading_result[0]  # Take first if multiple

            # Generate HTML tearsheet
            html_content = trading_result.to_html()

            # Save to temporary file
            with tempfile.NamedTemporaryFile(mode="w", suffix=".html", delete=False) as f:
                # Handle different types of HTML content and ensure it's a string
                if hasattr(html_content, "data"):
                    content_str = str(html_content.data)
                else:
                    content_str = str(html_content)
                f.write(content_str)
                temp_file_path = f.name

            # Open in browser
            webbrowser.open(f"file://{temp_file_path}")

            # Update the tearsheet viewer
            tearsheet_viewer = self.query_one("#tearsheet-info", TearsheetViewer)
            tearsheet_viewer.show_tearsheet_info(display_name, temp_file_path)

        except Exception as e:
            logger.error(f"Failed to generate tearsheet: {e}")
            tearsheet_viewer = self.query_one("#tearsheet-info", TearsheetViewer)
            tearsheet_viewer.update(f"Failed to generate tearsheet: {e}")


def run_backtest_browser(root_path: str):
    """Run the backtest browser TUI application"""
    try:
        app = BacktestBrowserApp(root_path)
        app.run()
    except Exception as e:
        logger.error(f"Failed to start backtest browser: {e}")
        raise



import csv
import os
from multiprocessing.pool import ThreadPool
from typing import Any

from qubx import logger
from qubx.core.loggers import LogsWriter
from qubx.utils.misc import makedirs


class CsvFileLogsWriter(LogsWriter):
    """
    Simple CSV strategy log data writer. It does data writing in separate thread.
    """

    def __init__(self, account_id: str, strategy_id: str, run_id: str, log_folder="logs") -> None:
        super().__init__(account_id, strategy_id, run_id)

        path = makedirs(log_folder)
        # - it rewrites positions every time
        self._pos_file_path = f"{path}/{self.strategy_id}_{self.account_id}_positions.csv"
        self._balance_file_path = f"{path}/{self.strategy_id}_{self.account_id}_balance.csv"

        _pfl_path = f"{path}/{strategy_id}_{account_id}_portfolio.csv"
        _exe_path = f"{path}/{strategy_id}_{account_id}_executions.csv"
        _sig_path = f"{path}/{strategy_id}_{account_id}_signals.csv"
        _tgt_path = f"{path}/{strategy_id}_{account_id}_targets.csv"

        self._hdr_pfl = not os.path.exists(_pfl_path)
        self._hdr_exe = not os.path.exists(_exe_path)
        self._hdr_sig = not os.path.exists(_sig_path)
        self._hdr_tgt = not os.path.exists(_tgt_path)

        self._pfl_file_ = open(_pfl_path, "a", newline="")
        self._execs_file_ = open(_exe_path, "a", newline="")
        self._sig_file_ = open(_sig_path, "a", newline="")
        self._tgt_file_ = open(_tgt_path, "a", newline="")

        self._pfl_writer = csv.writer(self._pfl_file_)
        self._exe_writer = csv.writer(self._execs_file_)
        self._sig_writer = csv.writer(self._sig_file_)
        self._tgt_writer = csv.writer(self._tgt_file_)

        self.pool = ThreadPool(4)

    @staticmethod
    def _header(d: dict) -> list[str]:
        return list(d.keys()) + ["run_id"]

    def _values(self, data: list[dict[str, Any]]) -> list[list[str]]:
        # - attach run_id (last column)
        return [list((d | {"run_id": self.run_id}).values()) for d in data]

    def _do_write(self, log_type: str, data: list[dict[str, Any]]):
        match log_type:
            case "positions":
                with open(self._pos_file_path, "w", newline="") as f:
                    w = csv.writer(f)
                    w.writerow(self._header(data[0]))
                    w.writerows(self._values(data))

            case "portfolio":
                if self._hdr_pfl:
                    self._pfl_writer.writerow(self._header(data[0]))
                    self._hdr_pfl = False
                self._pfl_writer.writerows(self._values(data))
                self._pfl_file_.flush()

            case "executions":
                if self._hdr_exe:
                    self._exe_writer.writerow(self._header(data[0]))
                    self._hdr_exe = False
                self._exe_writer.writerows(self._values(data))
                self._execs_file_.flush()

            case "signals":
                if self._hdr_sig:
                    self._sig_writer.writerow(self._header(data[0]))
                    self._hdr_sig = False
                self._sig_writer.writerows(self._values(data))
                self._sig_file_.flush()

            case "targets":
                if self._hdr_tgt:
                    self._tgt_writer.writerow(self._header(data[0]))
                    self._hdr_tgt = False
                self._tgt_writer.writerows(self._values(data))
                self._tgt_file_.flush()

            case "balance":
                with open(self._balance_file_path, "w", newline="") as f:
                    w = csv.writer(f)
                    w.writerow(self._header(data[0]))
                    w.writerows(self._values(data))

    def write_data(self, log_type: str, data: list[dict[str, Any]]):
        if len(data) > 0:
            self.pool.apply_async(self._do_write, (log_type, data))

    def flush_data(self):
        try:
            self._pfl_file_.flush()
            self._execs_file_.flush()
            self._sig_file_.flush()
            self._tgt_file_.flush()
        except Exception as e:
            logger.warning(f"Error flushing log writer: {str(e)}")

    def close(self):
        self._pfl_file_.close()
        self._execs_file_.close()
        self._sig_file_.close()
        self._tgt_file_.close()
        self.pool.close()
        self.pool.join()



# cython: language_level=3, boundscheck=False, wraparound=False, cdivision=True

cimport cython
import numpy as np
cimport numpy as np

from qubx.core.series cimport OrderBook
from qubx.utils.hft.numba_utils import argsort
from qubx.utils.orderbook import accumulate_orderbook_levels

cdef double[:, :] two_d_float_array(int depth):
    cdef np.ndarray arr = np.zeros((depth, 2), dtype=np.float64)
    return arr

cdef class LOB:
    """
    A class representing a Limit Order Book (LOB).

    Attributes:
        timestamp (np.int64_t): The current timestamp.
        depth (int): The depth of the order book.
        safe_depth (int): The safe depth of the order book.
        bids (double[:, :]): The bid levels.
        asks (double[:, :]): The ask levels.
        timestamps_buffer (RingBuffer): The buffer for timestamps.
        bids_buffer (RingBuffer): The buffer for bid levels.
        asks_buffer (RingBuffer): The buffer for ask levels.
        mids_buffer (RingBuffer): The buffer for mid prices.
    """
    cdef:
        public np.int64_t timestamp
        public int depth, safe_depth, buffer_size
        public double[:, :] bids, asks, bids_cache, asks_cache

    def __init__(
        self,
        np.int64_t timestamp=-1,
        double[:, :] bids=None,
        double[:, :] asks=None,
        int depth=100,
        bint apply_shadow_depth=False,
        **kwargs
    ):
        """
        Initialize the LOB.

        Args:
            timestamp (np.int64_t, optional): The initial timestamp in ms. Defaults to -1.
            bids (double[:, :], optional): The initial bid levels. Defaults to None.
            asks (double[:, :], optional): The initial ask levels. Defaults to None.
            depth (int, optional): The depth of the order book. Defaults to 100.
            apply_shadow_depth (bint, optional): Whether to maintain higher levels than depth for accuracy at speed cost. Defaults to False.
            **kwargs: Additional keyword arguments.
        """
        self.timestamp = timestamp 
        self.depth = depth 
        self.safe_depth = depth*3 if apply_shadow_depth else depth

        self.bids = two_d_float_array(self.safe_depth)
        self.asks = two_d_float_array(self.safe_depth)
        self.bids_cache = two_d_float_array(self.safe_depth)
        self.asks_cache = two_d_float_array(self.safe_depth)
        
        cdef int bid_idx, ask_idx, i
        
        if bids is not None:
            bids_np = np.asarray(bids)
            bids_np = bids_np[argsort(bids_np[:,0])][::-1]
            bid_idx = min(bids_np.shape[0], self.safe_depth)
            for i in range(bid_idx):
                self.bids[i, 0] = bids_np[i, 0]
                self.bids[i, 1] = bids_np[i, 1]
            self.bids[bid_idx:, 1] = 0
        
        if asks is not None:            
            asks_np = np.asarray(asks)
            asks_np = asks_np[argsort(asks_np[:,0])]
            ask_idx = min(asks_np.shape[0], self.safe_depth)
            for i in range(ask_idx):
                self.asks[i, 0] = asks_np[i, 0]
                self.asks[i, 1] = asks_np[i, 1]
            self.asks[ask_idx:, 1] = 0

    def update(
        self, np.int64_t timestamp, double[:, :] bids, double[:, :] asks, bint is_snapshot, bint is_sorted=True
    ):
        """
        Update the order book with new bid and ask levels.

        Args:
            timestamp (np.int64_t): The new timestamp.
            bids (double[:, :]): The new bid levels.
            asks (double[:, :]): The new ask levels.
            is_snapshot (bint): Whether the update is a snapshot or delta update.
            is_sorted (bint, optional): Whether the levels are already sorted (increasing/decreasing). Defaults to True.
        """
        cdef:
            bint bids_provided = bids is not None and bids.shape[0] != 0
            bint asks_provided = asks is not None and asks.shape[0] != 0
            bint bids_is_increasing, asks_is_increasing
            int i
            int bid_idx, ask_idx
            int old_idx, new_idx, cache_idx
            double old_price, old_volume, new_price, new_volume
            double[:, :] temp_view
            np.ndarray bids_np, asks_np
       
        if is_sorted:
            if bids_provided:
                bids_is_increasing = bids[0, 0] < bids[bids.shape[0]-1, 0]
                if bids_is_increasing:
                    bids_np = np.asarray(bids)[::-1]
                    bids = bids_np
            
            if asks_provided:
                asks_is_increasing = asks[0, 0] < asks[asks.shape[0]-1, 0]
                if not asks_is_increasing:
                    asks_np = np.asarray(asks)[::-1]
                    asks = asks_np
     
        if not is_sorted:
            if bids_provided:
                bids_np = np.asarray(bids)
                bids_np = bids_np[argsort(bids_np[:,0])][::-1]
                bids = bids_np
            if asks_provided:
                asks_np = np.asarray(asks)
                asks_np = asks_np[argsort(asks_np[:,0])]
                asks = asks_np
        
        if is_snapshot:
            if bids_provided:
                bid_idx = min(bids.shape[0], self.safe_depth)
                for i in range(bid_idx):
                    self.bids[i, :] = bids[i, :]
                self.bids[bid_idx:, 1] = 0
            
            if asks_provided:
                ask_idx = min(asks.shape[0], self.safe_depth)
                for i in range(ask_idx):
                    self.asks[i, :] = asks[i, :]
                self.asks[ask_idx:, 1] = 0
        else:
            if bids_provided:
                old_idx = 0
                new_idx = 0
                cache_idx = 0
                
                while old_idx < self.safe_depth and new_idx < bids.shape[0] and cache_idx < self.safe_depth:
                    old_price = self.bids[old_idx, 0]
                    old_volume = self.bids[old_idx, 1]
                    new_price = bids[new_idx, 0]
                    new_volume = bids[new_idx, 1]

                    if old_volume == 0:
                        old_idx = self.safe_depth
                        break
                        
                    if old_price > new_price:  
                        self.bids_cache[cache_idx,:] = self.bids[old_idx,:]
                        cache_idx += 1
                        old_idx += 1
                    elif old_price < new_price: 
                        if new_volume > 0:
                            self.bids_cache[cache_idx,:] = bids[new_idx,:]
                            cache_idx += 1
                        new_idx += 1
                    else: 
                        if new_volume > 0:
                            self.bids_cache[cache_idx,:] = bids[new_idx,:]
                            cache_idx += 1
                        old_idx += 1
                        new_idx += 1
               
                while old_idx < self.safe_depth and cache_idx < self.safe_depth:
                    if self.bids[old_idx, 1] == 0:
                        break
                    self.bids_cache[cache_idx,:] = self.bids[old_idx,:]
                    old_idx += 1
                    cache_idx += 1
                
                while new_idx < bids.shape[0] and cache_idx < self.safe_depth:
                    if bids[new_idx, 1] > 0:
                        self.bids_cache[cache_idx,:] = bids[new_idx,:]
                        cache_idx += 1
                    new_idx += 1
                  
                if cache_idx < self.safe_depth:
                    self.bids_cache[cache_idx:, 1] = 0
                
                temp_view = self.bids
                self.bids = self.bids_cache
                self.bids_cache = temp_view
            
            if asks_provided:
                old_idx = 0
                new_idx = 0
                cache_idx = 0
                
                while old_idx < self.safe_depth and new_idx < asks.shape[0] and cache_idx < self.safe_depth:
                    old_price = self.asks[old_idx, 0]
                    old_volume = self.asks[old_idx, 1]
                    new_price = asks[new_idx, 0]
                    new_volume = asks[new_idx, 1]
                    
                    if old_volume == 0:
                        old_idx = self.safe_depth
                        break
                        
                    if old_price < new_price:
                        self.asks_cache[cache_idx,:] = self.asks[old_idx,:]
                        cache_idx += 1
                        old_idx += 1
                    elif old_price > new_price:
                        if new_volume > 0:
                            self.asks_cache[cache_idx,:] = asks[new_idx,:]
                            cache_idx += 1
                        new_idx += 1
                    else: 
                        if new_volume > 0:
                            self.asks_cache[cache_idx,:] = asks[new_idx,:]
                            cache_idx += 1
                        old_idx += 1
                        new_idx += 1
                        
                while old_idx < self.safe_depth and cache_idx < self.safe_depth:
                    if self.asks[old_idx, 1] == 0:
                        break
                    self.asks_cache[cache_idx,:] = self.asks[old_idx,:]
                    old_idx += 1
                    cache_idx += 1
                
                while new_idx < asks.shape[0] and cache_idx < self.safe_depth:
                    if asks[new_idx, 1] > 0:
                        self.asks_cache[cache_idx,:] = asks[new_idx,:]
                        cache_idx += 1
                    new_idx += 1
                
                if cache_idx < self.safe_depth:
                    self.asks_cache[cache_idx:, 1] = 0
                
                temp_view = self.asks
                self.asks = self.asks_cache
                self.asks_cache = temp_view
        
        self.timestamp = timestamp

    cpdef double get_mid(self):
        """
        Get the mid price.

        Returns:
            double: The mid price.
        """
        cdef double best_bid = self.get_bid()
        cdef double best_ask = self.get_ask()
        return (best_bid + best_ask) / 2

    cpdef np.ndarray get_bids(self,copy=False):
        """
        Get the bid levels.

        Returns:
            np.ndarray: The bid levels.
        """
        bids = np.asarray(self.bids[:self.depth, :])
        return bids.copy() if copy else bids

    cpdef np.ndarray get_asks(self,copy=False):
        """
        Get the ask levels.

        Returns:
            np.ndarray: The ask levels.
        """
        asks = np.asarray(self.asks[:self.depth, :])
        return asks.copy() if copy else asks

    cpdef double get_bid(self):
        """
        Get the best bid price.

        Returns:
            double: The best bid price.
        """
        return self.bids[0, 0] if self.bids[0, 1] != 0 else np.nan

    cpdef double get_ask(self):
        """
        Get the best ask price.

        Returns:
            double: The best ask price.
        """
        return self.asks[0, 0] if self.asks[0, 1] != 0 else np.nan

    cpdef double get_bid_sz(self):
        """
        Get the size of the best bid.

        Returns:
            double: The size of the best bid.
        """
        return self.bids[0, 1]

    cpdef double get_ask_sz(self):
        """
        Get the size of the best ask.

        Returns:
            double: The size of the best ask.
        """
        return self.asks[0, 1]

    cpdef tuple as_tuple(self,copy=True):
        """
        Get the order book as a tuple.

        Returns:
            tuple: The order book as a tuple (timestamp, bids, asks).
        """
        return (self.timestamp, self.get_bids(copy=copy), self.get_asks(copy=copy))

    cpdef dict as_dict(self,copy=True):
        """
        Get the order book as a dictionary.

        Returns:
            dict: The order book as a dictionary.
        """
        return {
            "ts": self.timestamp,
            "b": self.get_bids(copy=copy),
            "a": self.get_asks(copy=copy)
        }

    cpdef OrderBook get_orderbook(self, double tick_size, int levels, bint sizes_in_quoted=False):
        """
        Generate OrderBook from current LOB state with tick-size aggregation.

        Args:
            tick_size: Price tick size for aggregation into uniform grid
            levels: Number of price levels to include in output
            sizes_in_quoted: Whether sizes should be in quoted currency (price * size)

        Returns:
            OrderBook object with aggregated levels, or None if state is empty/invalid

        Note:
            For levels==1, returns best bid/ask without aggregation for performance.
            For levels>1, aggregates raw orderbook into uniform tick-size grid using
            the accumulate_orderbook_levels function.
        """
        # Check if orderbook has data
        if self.bids[0, 1] == 0 or self.asks[0, 1] == 0:
            return None

        cdef:
            double best_bid = self.bids[0, 0]
            double best_ask = self.asks[0, 0]
            int actual_levels = min(levels, self.depth)
            np.ndarray sorted_bids, sorted_asks
            np.ndarray bids_buffer, asks_buffer
            double top_bid_agg, top_ask_agg

        # Detect crossed orderbook (should not happen with clean data)
        if best_bid >= best_ask:
            return None

        # Optimization: for depth==1, skip aggregation and return best bid/ask directly
        if levels == 1:
            return OrderBook(
                time=self.timestamp,
                top_bid=best_bid,
                top_ask=best_ask,
                tick_size=tick_size,
                bids=np.array([self.bids[0, 1]], dtype=np.float64),
                asks=np.array([self.asks[0, 1]], dtype=np.float64),
            )

        # Get bids/asks as numpy arrays (slice to requested levels)
        sorted_bids = np.asarray(self.bids[:actual_levels, :], dtype=np.float64)
        sorted_asks = np.asarray(self.asks[:actual_levels, :], dtype=np.float64)

        # Allocate buffers for aggregated levels
        bids_buffer = np.zeros(actual_levels, dtype=np.float64)
        asks_buffer = np.zeros(actual_levels, dtype=np.float64)

        # Apply tick-size aggregation using numba function
        # This accumulates raw levels into uniform price grid based on tick_size
        top_bid_agg, bids_buffer = accumulate_orderbook_levels(
            sorted_bids,
            bids_buffer,
            tick_size,
            True,  # is_bid
            actual_levels,
            sizes_in_quoted,
        )

        top_ask_agg, asks_buffer = accumulate_orderbook_levels(
            sorted_asks,
            asks_buffer,
            tick_size,
            False,  # is_bid
            actual_levels,
            sizes_in_quoted,
        )

        return OrderBook(
            time=self.timestamp,
            top_bid=top_bid_agg,
            top_ask=top_ask_agg,
            tick_size=tick_size,
            bids=bids_buffer,
            asks=asks_buffer,
        )



from collections.abc import Iterator
from typing import Any

import numpy as np
import pandas as pd

from qubx.core.basics import DataType
from qubx.core.series import OHLCV
from qubx.data.storage import IDataTransformer, Transformable
from qubx.data.storages.utils import find_time_col_idx
from qubx.data.transformers import OHLCVSeries, PandasFrame, TickSeries, TypedRecords


class TransformableWithHelpers(Transformable):
    """
    Handy transformer helpers to avoid typing x.transform(PandasFrame()) etc
    """

    def to_pd(self, id_in_index=False) -> pd.DataFrame:
        return self.transform(PandasFrame(id_in_index))

    def to_ohlc(self, timestamp_units="ns", max_length=np.inf) -> OHLCV:
        return self.transform(OHLCVSeries(timestamp_units, max_length))

    def to_ticks(
        self,
        trades: bool = False,  # if we also wants 'trades'
        default_bid_size=1e9,  # default bid/ask is big
        default_ask_size=1e9,  # default bid/ask is big
        daily_session_start_end=TickSeries.DEFAULT_DAILY_SESSION,
        timestamp_units="ns",
        spread=0.0,
        open_close_time_shift_secs=1.0,
        quotes=True,
    ) -> TypedRecords:
        return self.transform(
            TickSeries(
                trades,
                default_bid_size,
                default_ask_size,
                daily_session_start_end,
                timestamp_units,
                spread,
                open_close_time_shift_secs,
                quotes,
            )
        )

    def to_records(self, timestamp_units="ns") -> TypedRecords:
        return self.transform(TypedRecords(timestamp_units=timestamp_units))


class RawData(TransformableWithHelpers):
    """
    Container for raw market data from a single instrument/symbol.

    Holds untransformed data as returned by IReader.read() along with metadata needed
    for further processing. The raw data is stored as a list/array with named columns.

    Attributes:
        data_id: Symbol or instrument identifier (e.g., 'BTCUSDT', 'ETHUSDT')
        names: Column names from the data source (e.g., ['time', 'price', 'size'])
        dtype: Type of market data (DataType.TRADE, DataType.QUOTE, DataType.OHLC, etc.)
        raw: Raw data as numpy array or list of rows

    Usage:
        # - Read raw data from storage
        raw = reader.read("BTCUSDT", DataType.OHLC["1h"])

        # - Transform to pandas DataFrame
        df = raw.transform(PandasFrame())

        # - Transform to typed records (Trade, Quote, OHLCV objects)
        records = raw.transform(TypedRecords())

        # - Transform to OHLCV series
        ohlcv = raw.transform(OHLCVSeries())

        # - Transform to simulated tick data
        ticks = raw.transform(TickSeries(trades=True, quotes=True))

        # - Get time range
        start, end = raw.get_time_interval()

    The transform() method delegates to IDataTransformer implementations for
    converting raw data into various formats (DataFrames, typed objects, series, etc.).
    """

    data_id: str
    names: list[str]
    dtype: DataType
    raw: list
    _index: int

    def __init__(self, data_id: str, field_names: list[str], dtype: DataType, data: list):
        self.data_id = data_id
        self.names = field_names
        self.dtype = dtype
        self.raw = data
        self._index = find_time_col_idx(field_names)

    def __len__(self) -> int:
        return len(self.raw)

    def get_time_interval(self) -> tuple:
        """
        Returns start and end timestamp from raw data
        """
        return (self.raw[0][self._index], self.raw[-1][self._index]) if len(self) > 0 else (None, None)

    def transform(self, transformer: IDataTransformer) -> Any:
        return transformer.process_data(self.data_id, self.dtype, self.raw, self.names, self._index)

    def __repr__(self) -> str:
        s, e = self.get_time_interval()
        _range = f"{s} : {e}" if s and e else "EMPTY"
        return f"{self.data_id}({self.dtype})[{_range}]"


class RawMultiData(TransformableWithHelpers):
    """
    Container for raw market data from multiple instruments/symbols.

    Aggregates multiple RawData instances for batch processing and multi-instrument
    transformations. All contained RawData must have the same underlying data type
    (same raw data structure) to ensure consistent processing.

    Attributes:
        raws: Dictionary mapping data_id (symbol) to RawData instances

    Usage:
        # - Read data for multiple symbols
        btc_raw = reader.read("BTCUSDT", DataType.OHLC["1h"])
        eth_raw = reader.read("ETHUSDT", DataType.OHLC["1h"])

        # - Combine into multi-data container
        multi = RawMultiData([btc_raw, eth_raw])

        # - Transform to multi-index DataFrame (symbol as level)
        df = multi.transform(PandasFrame(symbol_as_index=True))
        # Returns DataFrame with MultiIndex (timestamp, symbol)

        # - Transform to dict of DataFrames
        dfs = multi.transform(PandasFrame(symbol_as_index=False))
        # Returns {'BTCUSDT': df1, 'ETHUSDT': df2}


    The transform() method applies the transformer to each contained RawData and
    then calls combine_data() to merge results according to transformer logic.

    Raises:
        ValueError: If attempting to add RawData with incompatible raw data type
    """

    raws: dict[str, RawData]

    def __init__(self, data: list[RawData]):
        self.raws = {}

        _t = None
        for r in data:
            if not _t:
                _t = type(r.raw)
            elif _t is not type(r.raw):
                raise ValueError(f"RawMultiData container may contain only single data type {_t.__name__}")

            self.raws[r.data_id] = r

    def pop(self, data_id: str) -> RawData | None:
        if data_id in self.raws:
            return self.raws.pop(data_id)
        return None

    def add(self, r: RawData):
        self.raws[r.data_id] = r

    def get_time_interval(self, data_id: str) -> tuple:
        return self.raws[data_id].get_time_interval()

    def get_data_ids(self) -> list[str]:
        return list(self.raws.keys())

    def transform(self, transformer: IDataTransformer) -> Any:
        return transformer.combine_data({k: r.transform(transformer) for k, r in self.raws.items()})

    def __getitem__(self, data_id: str) -> RawData:
        return self.raws[data_id]

    def __len__(self) -> int:
        return len(self.raws)

    def __repr__(self) -> str:
        return "-[MULTI DATA]-\n" + "\n".join(["\t - " + repr(s) for s in self.raws.values()])


class IteratorsMaster:
    """
    Manages multiple iterators and advances them in parallel.
    """

    iterators: list[Iterator]

    def __init__(self, iterators: list[Iterator]):
        self.iterators = iterators

    def __iter__(self):
        return self

    def __next__(self) -> RawMultiData:
        """
        Advance all iterators and return their values as a list.

        Returns:
            List of values in the same order as the dictionary keys.

        Raises:
            StopIteration: When all underlying iterators are exhausted.
        """
        result = []
        all_exhausted = True

        for iter in self.iterators:
            try:
                value = next(iter)
                result.append(value)
                all_exhausted = False
            except StopIteration:
                pass

        if all_exhausted:
            raise StopIteration

        # - return multi data
        return RawMultiData(result)
